#-*- coding:utf-8 -*-

import unittest
from contextlib import contextmanager
from datetime import datetime
from time import sleep

import mock
from tornado import gen
from tornado import testing

from pybreaker import *


class CircuitBreakerStorageBasedTestCase(object):
    """
    Mix in to test against different storage backings. Depends on
    `self.breaker` and `self.breaker_kwargs`.
    """

    def test_successful_call(self):
        """CircuitBreaker: it should keep the circuit closed after a successful
        call.
        """
        def func(): return True
        self.assertTrue(self.breaker.call(func))
        self.assertEqual(1, self.breaker.success_counter)
        self.assertEqual(0, self.breaker.fail_counter)
        self.assertEqual('closed', self.breaker.current_state)

    def test_one_failed_call(self):
        """CircuitBreaker: it should keep the circuit closed after a few
        failures.
        """
        def func(): raise NotImplementedError()
        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertEqual(1, self.breaker.fail_counter)
        self.assertEqual('closed', self.breaker.current_state)

    def test_one_successful_call_after_failed_call(self):
        """CircuitBreaker: it should keep the circuit closed after few mixed
        outcomes.
        """
        def suc(): return True
        def err(): raise NotImplementedError()

        self.assertRaises(NotImplementedError, self.breaker.call, err)
        self.assertEqual(1, self.breaker.fail_counter)

        self.assertTrue(self.breaker.call(suc))
        self.assertEqual(1, self.breaker.success_counter)

        # as per the new change, this will not fall to '0', if one call succeed
        # self.assertEqual(0, self.breaker.fail_counter)

        self.assertEqual('closed', self.breaker.current_state)

    def test_several_failed_calls(self):
        """CircuitBreaker: it should open the circuit after many failures.
        """
        def func(): raise NotImplementedError()

        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertRaises(NotImplementedError, self.breaker.call, func)

        # Circuit should open
        self.assertRaises(CircuitBreakerError, self.breaker.call, func)
        self.assertEqual(5, self.breaker.fail_counter)
        self.assertEqual('open', self.breaker.current_state)

    def test_traceback_in_circuitbreaker_error(self):
        """CircuitBreaker: it should open the circuit after many failures.
        """
        def func(): raise NotImplementedError()

        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertRaises(NotImplementedError, self.breaker.call, func)

        # Circuit should open
        try:
            self.breaker.call(func)
            self.fail('CircuitBreakerError should throw')
        except CircuitBreakerError as e:
            import traceback
            self.assertIn('NotImplementedError', traceback.format_exc())
        self.assertEqual(5, self.breaker.fail_counter)
        self.assertEqual('open', self.breaker.current_state)

    def test_failed_call_after_timeout(self):
        """CircuitBreaker: it should half-open the circuit after timeout.
        """
        self.breaker = CircuitBreaker(reset_timeout=0.5, **self.breaker_kwargs)
        def func(): raise NotImplementedError()

        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertEqual('closed', self.breaker.current_state)
        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertRaises(NotImplementedError, self.breaker.call, func)

        # Circuit should open
        self.assertRaises(CircuitBreakerError, self.breaker.call, func)
        self.assertEqual(5, self.breaker.fail_counter)

        # Wait for timeout
        sleep(0.6)

        # Circuit should open again
        self.assertRaises(CircuitBreakerError, self.breaker.call, func)
        self.assertEqual(6, self.breaker.fail_counter)
        self.assertEqual('open', self.breaker.current_state)

    def test_successful_after_timeout(self):
        """CircuitBreaker: it should close the circuit when a call succeeds
        after timeout. The successful function should only be called once.
        """

        suc = mock.MagicMock(return_value=True)
        def err(): raise NotImplementedError()

        self.assertRaises(NotImplementedError, self.breaker.call, err)
        self.assertRaises(NotImplementedError, self.breaker.call, err)
        self.assertRaises(NotImplementedError, self.breaker.call, err)
        self.assertRaises(NotImplementedError, self.breaker.call, err)
        self.assertEqual('closed', self.breaker.current_state)

        # Circuit should open
        self.assertRaises(CircuitBreakerError, self.breaker.call, err)
        self.assertRaises(CircuitBreakerError, self.breaker.call, suc)
        self.assertEqual(5, self.breaker.fail_counter)

        self.breaker.reset_timeout = 3
        sleep(4)

        # Circuit should close again
        self.assertTrue(self.breaker.call(suc))
        self.assertEqual(1, self.breaker.success_counter)
        self.assertEqual(4, self.breaker.fail_counter)
        self.assertEqual('closed', self.breaker.current_state)
        self.assertEqual(1, suc.call_count)

    def test_failed_call_when_halfopen(self):
        """CircuitBreaker: it should open the circuit when a call fails in
        half-open state.
        """
        def fun(): raise NotImplementedError()

        self.breaker.half_open()
        self.assertEqual(0, self.breaker.fail_counter)
        self.assertEqual('half-open', self.breaker.current_state)

        # Circuit should open
        self.assertRaises(CircuitBreakerError, self.breaker.call, fun)
        self.assertEqual(1, self.breaker.fail_counter)
        self.assertEqual('open', self.breaker.current_state)

    def test_successful_call_when_halfopen(self):
        """CircuitBreaker: it should close the circuit when a call succeeds in
        half-open state.
        """
        def fun(): return True

        self.breaker.half_open()
        self.assertEqual(0, self.breaker.fail_counter)
        self.assertEqual('half-open', self.breaker.current_state)

        # Circuit should open
        self.assertTrue(self.breaker.call(fun))
        self.assertEqual(1, self.breaker.success_counter)
        self.assertEqual(0, self.breaker.fail_counter)
        self.assertEqual('closed', self.breaker.current_state)

    def test_call_when_total_calls_less_than_request_volume_window(self):
        """CircuitBreaker: it should not open the circuit until total calls
         are less than defined request_volume_window size.
        """

        def func(): raise NotImplementedError()

        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertRaises(NotImplementedError, self.breaker.call, func)

        self.assertEqual(4, self.breaker.fail_counter)
        self.assertEqual(0, self.breaker.success_counter)

        self.assertEqual('closed', self.breaker.current_state)

    def test_call_when_total_calls_greater_than_request_volume_window(self):
        """CircuitBreaker: it should not open the circuit until total calls
         are less than defined request_volume_window size.
        """

        def func(): raise NotImplementedError()

        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertRaises(NotImplementedError, self.breaker.call, func)

        self.assertRaises(CircuitBreakerError, self.breaker.call, func)

        self.assertEqual('open', self.breaker.current_state)

        self.assertEqual(5, self.breaker.fail_counter)
        self.assertEqual(0, self.breaker.success_counter)

        # number of failures greater than window size, as err_rate = 6/5 is greater than err_threshold, it will through CB error
        self.assertRaises(CircuitBreakerError, self.breaker.call, func)

    def test_close(self):
        """CircuitBreaker: it should allow the circuit to be closed manually.
        Now, it supports error rate so 3/5 (fail_call/total_calls)= 0.15
        """
        self.breaker = CircuitBreaker(**self.breaker_kwargs)
        def func1(): raise NotImplementedError()
        def func2(): return True

        self.assertRaises(NotImplementedError, self.breaker.call, func1)
        self.assertRaises(NotImplementedError, self.breaker.call, func1)
        self.assertRaises(NotImplementedError, self.breaker.call, func1)
        self.assertRaises(NotImplementedError, self.breaker.call, func1)
        self.assertRaises(CircuitBreakerError, self.breaker.call, func1)

        self.assertEqual(5, self.breaker.fail_counter)
        self.assertEqual('open', self.breaker.current_state)

        # Circuit should close again
        self.breaker.close()
        self.assertEqual('closed', self.breaker.current_state)

    def test_transition_events(self):
        """CircuitBreaker: it should call the appropriate functions on every
        state transition.
        """
        class Listener(CircuitBreakerListener):
            def __init__(self):
                self.out = ''

            def state_change(self, cb, old_state, new_state):
                assert cb
                if old_state: self.out += old_state.name
                if new_state: self.out += '->' + new_state.name
                self.out += ','

        listener = Listener()
        self.breaker = CircuitBreaker(listeners=(listener,), **self.breaker_kwargs)
        self.assertEqual('closed', self.breaker.current_state)

        self.breaker.open()
        self.assertEqual('open', self.breaker.current_state)

        self.breaker.half_open()
        self.assertEqual('half-open', self.breaker.current_state)

        self.breaker.close()
        self.assertEqual('closed', self.breaker.current_state)

        self.assertEqual('closed->open,open->half-open,half-open->closed,', \
                         listener.out)

    def test_call_events(self):
        """CircuitBreaker: it should call the appropriate functions on every
        successful/failed call.
        """
        self.out = ''

        def suc(): return True
        def err(): raise NotImplementedError()

        class Listener(CircuitBreakerListener):
            def __init__(self):
                self.out = ''
            def before_call(self, cb, func, *args, **kwargs):
                assert cb
                self.out += '-'
            def success(self, cb):
                assert cb
                self.out += 'success'
            def failure(self, cb, exc):
                assert cb; assert exc
                self.out += 'failure'

        listener = Listener()
        self.breaker = CircuitBreaker(listeners=(listener,), **self.breaker_kwargs)

        self.assertTrue(self.breaker.call(suc))
        self.assertRaises(NotImplementedError, self.breaker.call, err)
        self.assertEqual('-success-failure', listener.out)

    def test_generator(self):
        """CircuitBreaker: it should inspect generator values.
        """
        @self.breaker
        def suc(value):
            "Docstring"
            yield value

        @self.breaker
        def err(value):
            "Docstring"
            x = yield value
            raise NotImplementedError(x)

        s = suc(True)
        e = err(True)
        next(e)

        self.assertRaises(NotImplementedError, e.send, True)
        self.assertEqual(1, self.breaker.fail_counter)
        self.assertTrue(next(s))
        self.assertRaises((StopIteration, RuntimeError), lambda: next(s))

        # as per the new change, this will not fall to '0', if one call succeed
        # self.assertEqual(0, self.breaker.fail_counter)

    def test_contextmanager(self):
        """CircuitBreaker: it should catch in a with statement
        """
        class Foo:
            @contextmanager
            @self.breaker
            def wrapper(self):
                try:
                    yield
                except NotImplementedError as e:
                    raise ValueError()

            def foo(self):
                with self.wrapper():
                    raise NotImplementedError()


        try:
            Foo().foo()
        except ValueError as e:
            self.assertTrue(isinstance(e, ValueError))


class CircuitBreakerConfigurationTestCase(object):
    """
    Tests for the CircuitBreaker class.
    """

    def test_default_state(self):
        """CircuitBreaker: it should get initial state from state_storage.
        """
        for state in (STATE_OPEN, STATE_CLOSED, STATE_HALF_OPEN):
            storage = CircuitMemoryStorage(state)
            breaker = CircuitBreaker(state_storage=storage)
            self.assertEqual(breaker.state.name, state)

    def test_default_params(self):
        """CircuitBreaker: it should define smart defaults.
        """
        self.assertEqual(0, self.breaker.fail_counter)
        self.assertEqual(30, self.breaker.reset_timeout)
        self.assertEqual('closed', self.breaker.current_state)
        self.assertEqual((), self.breaker.excluded_exceptions)
        self.assertEqual((), self.breaker.listeners)
        self.assertEqual('memory', self.breaker._state_storage.name)

    def test_new_with_custom_reset_timeout(self):
        """CircuitBreaker: it should support a custom reset timeout value.
        """
        self.breaker = CircuitBreaker(reset_timeout=30)
        self.assertEqual(0, self.breaker.fail_counter)
        self.assertEqual(30, self.breaker.reset_timeout)
        self.assertEqual((), self.breaker.excluded_exceptions)
        self.assertEqual((), self.breaker.listeners)
        self.assertEqual('memory', self.breaker._state_storage.name)

    def test_new_with_custom_fail_max(self):
        """CircuitBreaker: it should support a custom maximum number of
        failures.
        """
        self.assertEqual(0, self.breaker.fail_counter)
        self.assertEqual(30, self.breaker.reset_timeout)
        self.assertEqual((), self.breaker.excluded_exceptions)
        self.assertEqual((), self.breaker.listeners)
        self.assertEqual('memory', self.breaker._state_storage.name)

    def test_new_with_custom_excluded_exceptions(self):
        """CircuitBreaker: it should support a custom list of excluded
        exceptions.
        """
        self.breaker = CircuitBreaker(exclude=[Exception])
        self.assertEqual(0, self.breaker.fail_counter)
        self.assertEqual(30, self.breaker.reset_timeout)
        self.assertEqual((Exception,), self.breaker.excluded_exceptions)
        self.assertEqual((), self.breaker.listeners)
        self.assertEqual('memory', self.breaker._state_storage.name)

    def test_err_threshold_setter(self):
        """CircuitBreaker: it should allow the user to set a new value for
        'fail_max'.
        """
        self.assertEqual(0.2, self.breaker.err_threshold)
        self.breaker.err_threshold = 0.5
        self.assertEqual(0.5, self.breaker.err_threshold)

    def test_reset_timeout_setter(self):
        """CircuitBreaker: it should allow the user to set a new value for
        'reset_timeout'.
        """
        self.assertEqual(30, self.breaker.reset_timeout)
        self.breaker.reset_timeout = 20
        self.assertEqual(20, self.breaker.reset_timeout)

    def test_call_with_no_args(self):
        """CircuitBreaker: it should be able to invoke functions with no-args.
        """
        def func(): return True
        self.assertTrue(self.breaker.call(func))

    def test_call_with_args(self):
        """CircuitBreaker: it should be able to invoke functions with args.
        """
        def func(arg1, arg2): return [arg1, arg2]
        self.assertEqual([42, 'abc'], self.breaker.call(func, 42, 'abc'))

    def test_call_with_kwargs(self):
        """CircuitBreaker: it should be able to invoke functions with kwargs.
        """
        def func(**kwargs): return kwargs
        self.assertEqual({'a':1, 'b':2}, self.breaker.call(func, a=1, b=2))

    @testing.gen_test
    def test_call_async_with_no_args(self):
        """CircuitBreaker: it should be able to invoke async functions with no-args.
        """
        @gen.coroutine
        def func(): return True

        ret = yield self.breaker.call(func)
        self.assertTrue(ret)

    @testing.gen_test
    def test_call_async_with_args(self):
        """CircuitBreaker: it should be able to invoke async functions with args.
        """
        @gen.coroutine
        def func(arg1, arg2): return [arg1, arg2]

        ret = yield self.breaker.call(func, 42, 'abc')
        self.assertEqual([42, 'abc'], ret)

    @testing.gen_test
    def test_call_async_with_kwargs(self):
        """CircuitBreaker: it should be able to invoke async functions with kwargs.
        """
        @gen.coroutine
        def func(**kwargs): return kwargs
        ret = yield self.breaker.call(func, a=1, b=2)
        self.assertEqual({'a':1, 'b':2}, ret)

    def test_add_listener(self):
        """CircuitBreaker: it should allow the user to add a listener at a
        later time.
        """
        self.assertEqual((), self.breaker.listeners)

        first = CircuitBreakerListener()
        self.breaker.add_listener(first)
        self.assertEqual((first,), self.breaker.listeners)

        second = CircuitBreakerListener()
        self.breaker.add_listener(second)
        self.assertEqual((first, second), self.breaker.listeners)

    def test_add_listeners(self):
        """CircuitBreaker: it should allow the user to add listeners at a
        later time.
        """
        first, second = CircuitBreakerListener(), CircuitBreakerListener()
        self.breaker.add_listeners(first, second)
        self.assertEqual((first, second), self.breaker.listeners)

    def test_remove_listener(self):
        """CircuitBreaker: it should allow the user to remove a listener.
        """
        first = CircuitBreakerListener()
        self.breaker.add_listener(first)
        self.assertEqual((first,), self.breaker.listeners)

        self.breaker.remove_listener(first)
        self.assertEqual((), self.breaker.listeners)

    def test_excluded_exceptions(self):
        """CircuitBreaker: it should ignore specific exceptions.
        """
        self.breaker = CircuitBreaker(exclude=[LookupError])

        def err_1(): raise NotImplementedError()
        def err_2(): raise LookupError()
        def err_3(): raise KeyError()

        self.assertRaises(NotImplementedError, self.breaker.call, err_1)
        self.assertEqual(1, self.breaker.fail_counter)
        self.assertEqual(0, self.breaker.success_counter)

        # LookupError is not considered a system error
        self.assertRaises(LookupError, self.breaker.call, err_2)
        self.assertEqual(1, self.breaker.fail_counter)
        self.assertEqual(1, self.breaker.success_counter)

        self.assertRaises(NotImplementedError, self.breaker.call, err_1)
        self.assertEqual(2, self.breaker.fail_counter)
        self.assertEqual(1, self.breaker.success_counter)

        # Should consider subclasses as well (KeyError is a subclass of
        # LookupError)
        self.assertRaises(KeyError, self.breaker.call, err_3)
        self.assertEqual(2, self.breaker.fail_counter)

    def test_excluded_callable_exceptions(self):
        """CircuitBreaker: it should ignore specific exceptions that return true from a filtering callable.
        """
        class TestException(Exception):
            def __init__(self, value):
                self.value = value

        filter_function = lambda e: type(e) == TestException and e.value == 'good'
        self.breaker = CircuitBreaker(exclude=[filter_function])

        def err_1(): raise TestException("bad")
        def err_2(): raise TestException("good")
        def err_3(): raise NotImplementedError()

        self.assertRaises(TestException, self.breaker.call, err_1)
        self.assertEqual(1, self.breaker.fail_counter)
        self.assertEqual(0, self.breaker.success_counter)

        self.assertRaises(TestException, self.breaker.call, err_2)
        self.assertEqual(1, self.breaker.fail_counter)
        self.assertEqual(1, self.breaker.success_counter)

        self.assertRaises(NotImplementedError, self.breaker.call, err_3)
        self.assertEqual(2, self.breaker.fail_counter)
        self.assertEqual(1, self.breaker.success_counter)

    def test_excluded_callable_and_types_exceptions(self):
        """CircuitBreaker: it should allow a mix of exclusions that includes both filter functions and types.
        """
        class TestException(Exception):
            def __init__(self, value):
                self.value = value

        filter_function = lambda e: type(e) == TestException and e.value == 'good'
        self.breaker = CircuitBreaker(exclude=[filter_function, LookupError])

        def err_1(): raise TestException("bad")
        def err_2(): raise TestException("good")
        def err_3(): raise NotImplementedError()
        def err_4(): raise LookupError()

        self.assertRaises(TestException, self.breaker.call, err_1)
        self.assertEqual(1, self.breaker.fail_counter)

        self.assertRaises(TestException, self.breaker.call, err_2)
        self.assertEqual(1, self.breaker.fail_counter)
        self.assertEqual(1, self.breaker.success_counter)

        self.assertRaises(NotImplementedError, self.breaker.call, err_3)
        self.assertEqual(2, self.breaker.fail_counter)
        self.assertEqual(1, self.breaker.success_counter)

        self.assertRaises(LookupError, self.breaker.call, err_4)
        self.assertEqual(2, self.breaker.fail_counter)
        self.assertEqual(2, self.breaker.success_counter)

    def test_add_excluded_exception(self):
        """CircuitBreaker: it should allow the user to exclude an exception at a
        later time.
        """
        self.assertEqual((), self.breaker.excluded_exceptions)

        self.breaker.add_excluded_exception(NotImplementedError)
        self.assertEqual((NotImplementedError,), \
                         self.breaker.excluded_exceptions)

        self.breaker.add_excluded_exception(Exception)
        self.assertEqual((NotImplementedError, Exception), \
                         self.breaker.excluded_exceptions)

    def test_add_excluded_exceptions(self):
        """CircuitBreaker: it should allow the user to exclude exceptions at a
        later time.
        """
        self.breaker.add_excluded_exceptions(NotImplementedError, Exception)
        self.assertEqual((NotImplementedError, Exception), \
                         self.breaker.excluded_exceptions)

    def test_remove_excluded_exception(self):
        """CircuitBreaker: it should allow the user to remove an excluded
        exception.
        """
        self.breaker.add_excluded_exception(NotImplementedError)
        self.assertEqual((NotImplementedError,), \
                         self.breaker.excluded_exceptions)

        self.breaker.remove_excluded_exception(NotImplementedError)
        self.assertEqual((), self.breaker.excluded_exceptions)

    def test_decorator(self):
        """CircuitBreaker: it should be a decorator.
        """
        @self.breaker
        def suc(value):
            "Docstring"
            return value

        @self.breaker
        def err(value):
            "Docstring"
            raise NotImplementedError()

        self.assertEqual('Docstring', suc.__doc__)
        self.assertEqual('Docstring', err.__doc__)
        self.assertEqual('suc', suc.__name__)
        self.assertEqual('err', err.__name__)

        self.assertRaises(NotImplementedError, err, True)
        self.assertEqual(1, self.breaker.fail_counter)

        self.assertTrue(suc(True))
        self.assertEqual(1, self.breaker.success_counter)

    @testing.gen_test
    def test_decorator_call_future(self):
        """CircuitBreaker: it should be a decorator.
        """
        @self.breaker(__pybreaker_call_async=True)
        @gen.coroutine
        def suc(value):
            "Docstring"
            raise gen.Return(value)

        @self.breaker(__pybreaker_call_async=True)
        @gen.coroutine
        def err(value):
            "Docstring"
            raise NotImplementedError()

        self.assertEqual('Docstring', suc.__doc__)
        self.assertEqual('Docstring', err.__doc__)
        self.assertEqual('suc', suc.__name__)
        self.assertEqual('err', err.__name__)

        with self.assertRaises(NotImplementedError):
            yield err(True)

        self.assertEqual(1, self.breaker.fail_counter)

        ret = yield suc(True)
        self.assertTrue(ret)
        self.assertEqual(1, self.breaker.fail_counter)
        self.assertEqual(1, self.breaker.success_counter)

    @mock.patch('pybreaker.HAS_TORNADO_SUPPORT', False)
    def test_no_tornado_raises(self):
        with self.assertRaises(ImportError):
            def func(): return True
            self.breaker(func, __pybreaker_call_async=True)

    def test_name(self):
        """CircuitBreaker: it should allow an optional name to be set and
           retrieved.
        """
        name = "test_breaker"
        self.breaker = CircuitBreaker(name=name)
        self.assertEqual(self.breaker.name, name)

        name = "breaker_test"
        self.breaker.name = name
        self.assertEqual(self.breaker.name, name)


class CircuitBreakerTestCase(testing.AsyncTestCase, CircuitBreakerStorageBasedTestCase, CircuitBreakerConfigurationTestCase):
    """
    Tests for the CircuitBreaker class.
    """

    def setUp(self):
        super(CircuitBreakerTestCase, self).setUp()
        self.breaker_kwargs = {}
        self.breaker = CircuitBreaker()

    def test_create_new_state__bad_state(self):
        with self.assertRaises(ValueError):
            self.breaker._create_new_state('foo')

    @mock.patch('pybreaker.CircuitOpenState')
    def test_notify_not_called_on_init(self, open_state):
        storage = CircuitMemoryStorage('open')
        breaker = CircuitBreaker(state_storage=storage)
        open_state.assert_called_once_with(breaker, prev_state=None, notify=False)

    @mock.patch('pybreaker.CircuitOpenState')
    def test_notify_called_on_state_change(self, open_state):
        storage = CircuitMemoryStorage('closed')
        breaker = CircuitBreaker(state_storage=storage)
        prev_state = breaker.state
        breaker.state = 'open'
        open_state.assert_called_once_with(breaker, prev_state=prev_state, notify=True)

    def test_failure_count_not_reset_during_creation(self):
        for state in (STATE_OPEN, STATE_CLOSED, STATE_HALF_OPEN):
            storage = CircuitMemoryStorage(state)
            storage.increment_counter_failure()
            #storage.increment_counter_success()

            breaker = CircuitBreaker(state_storage=storage)
            self.assertEqual(breaker.state.name, state)
            self.assertEqual(breaker.fail_counter, 1)

    def test_state_opened_at_not_reset_during_creation(self):
        for state in (STATE_OPEN, STATE_CLOSED, STATE_HALF_OPEN):
            storage = CircuitMemoryStorage(state)
            now = datetime.now()
            storage.opened_at = now

            breaker = CircuitBreaker(state_storage=storage)
            self.assertEqual(breaker.state.name, state)
            self.assertEqual(storage.opened_at, now)



#import fakeredis
import logging
from redis.exceptions import RedisError

class CircuitBreakerRedisTestCase(unittest.TestCase, CircuitBreakerStorageBasedTestCase):
    """
    Tests for the CircuitBreaker class.
    """

    def setUp(self):
        self.redis = fakeredis.FakeStrictRedis()
        self.breaker_kwargs = {'state_storage': CircuitRedisStorage('closed', self.redis)}
        self.breaker = CircuitBreaker(**self.breaker_kwargs)

    def tearDown(self):
        self.redis.flushall()

    def test_namespace(self):
        self.redis.flushall()
        self.breaker_kwargs = {'state_storage': CircuitRedisStorage('closed', self.redis, namespace='my_app')}
        self.breaker = CircuitBreaker(**self.breaker_kwargs)

        def func(): raise NotImplementedError()
        self.assertRaises(NotImplementedError, self.breaker.call, func)
        keys = self.redis.keys()
        self.assertEqual(2, len(keys))
        self.assertTrue(keys[0].decode('utf-8').startswith('my_app'))
        self.assertTrue(keys[1].decode('utf-8').startswith('my_app'))

    def test_fallback_state(self):
        logger = logging.getLogger('pybreaker')
        logger.setLevel(logging.FATAL)
        self.breaker_kwargs = {'state_storage': CircuitRedisStorage('closed', self.redis, fallback_circuit_state='open')}
        self.breaker = CircuitBreaker(**self.breaker_kwargs)
        def func(k): raise RedisError()
        with mock.patch.object(self.redis, 'get', new=func):
            state = self.breaker.state
            self.assertEqual('open', state.name)

    def test_missing_state(self):
        """CircuitBreakerRedis: If state on Redis is missing, it should set the
        fallback circuit state and reset the fail counter to 0.
        """
        self.breaker_kwargs = {'state_storage': CircuitRedisStorage('closed', self.redis, fallback_circuit_state='open')}
        self.breaker = CircuitBreaker(**self.breaker_kwargs)

        def func(): raise NotImplementedError()
        self.assertRaises(NotImplementedError, self.breaker.call, func)
        self.assertEqual(1, self.breaker.fail_counter)

        with mock.patch.object(self.redis, 'get', new=lambda k: None):
            state = self.breaker.state
            self.assertEqual('open', state.name)
            self.assertEqual(0, self.breaker.fail_counter)



import threading
from types import MethodType

class CircuitBreakerThreadsTestCase(unittest.TestCase):
    """
    Tests to reproduce common synchronization errors on CircuitBreaker class.
    """

    def setUp(self):
        self.breaker = CircuitBreaker(fail_max=3000, reset_timeout=1)

    def _start_threads(self, target, n):
        """
        Starts `n` threads that calls `target` and waits for them to finish.
        """
        threads = [threading.Thread(target=target) for i in range(n)]
        [t.start() for t in threads]
        [t.join() for t in threads]

    def _mock_function(self, obj, func):
        """
        Replaces a bounded function in `self.breaker` by another.
        """
        setattr(obj, func.__name__, MethodType(func, self.breaker))

    def test_fail_thread_safety(self):
        """CircuitBreaker: it should compute a failed call atomically to
        avoid race conditions.
        """
        # Create a specific exception to avoid masking other errors
        class SpecificException(Exception):
            pass

        @self.breaker
        def err(): raise SpecificException()

        def trigger_error():
            for n in range(500):
                try: err()
                except SpecificException: pass

        def _inc_counter(self):
            c = self._state_storage._fail_counter
            sleep(0.00005)
            self._state_storage._fail_counter = c + 1

        self._mock_function(self.breaker, _inc_counter)
        self._start_threads(trigger_error, 3)
        self.assertEqual(1500, self.breaker.fail_counter)

    def test_success_thread_safety(self):
        """CircuitBreaker: it should compute a successful call atomically
        to avoid race conditions.
        """
        @self.breaker
        def suc(): return True

        def trigger_success():
            for n in range(500):
                suc()

        class SuccessListener(CircuitBreakerListener):
            def success(self, cb):
                c = 0
                if hasattr(cb, '_success_counter'):
                    c = cb._success_counter
                sleep(0.00005)
                cb._success_counter = c + 1

        self.breaker.add_listener(SuccessListener())
        self._start_threads(trigger_success, 3)
        self.assertEqual(1500, self.breaker._success_counter)

    def test_half_open_thread_safety(self):
        """CircuitBreaker: it should allow only one trial call when the
        circuit is half-open.
        """
        self.breaker = CircuitBreaker(fail_max=1, reset_timeout=0.01)

        self.breaker.open()
        sleep(0.01)

        @self.breaker
        def err(): raise Exception()

        def trigger_failure():
            try: err()
            except: pass

        class StateListener(CircuitBreakerListener):
            def __init__(self):
                self._count = 0

            def before_call(self, cb, fun, *args, **kwargs):
                sleep(0.00005)

            def state_change(self, cb, old_state, new_state):
                if new_state.name == 'half-open':
                    self._count += 1

        state_listener = StateListener()
        self.breaker.add_listener(state_listener)

        self._start_threads(trigger_failure, 5)
        self.assertEqual(1, state_listener._count)


    def test_fail_max_thread_safety(self):
        """CircuitBreaker: it should not allow more failed calls than
        'fail_max' setting.
        """
        @self.breaker
        def err(): raise Exception()

        def trigger_error():
            for i in range(2000):
                try: err()
                except: pass

        class SleepListener(CircuitBreakerListener):
            def before_call(self, cb, func, *args, **kwargs):
                sleep(0.00005)

        self.breaker.add_listener(SleepListener())
        self._start_threads(trigger_error, 3)
        self.assertEqual(self.breaker.fail_max, self.breaker.fail_counter)


class CircuitBreakerRedisConcurrencyTestCase(unittest.TestCase):
    """
    Tests to reproduce common concurrency between different machines
    connecting to redis. This is simulated locally using threads.
    """

    def setUp(self):
        self.redis = fakeredis.FakeStrictRedis()
        self.breaker_kwargs = {'fail_max': 3000, 'reset_timeout': 1,'state_storage': CircuitRedisStorage('closed', self.redis)}
        self.breaker = CircuitBreaker(**self.breaker_kwargs)

    def tearDown(self):
        self.redis.flushall()

    def _start_threads(self, target, n):
        """
        Starts `n` threads that calls `target` and waits for them to finish.
        """
        threads = [threading.Thread(target=target) for i in range(n)]
        [t.start() for t in threads]
        [t.join() for t in threads]

    def _mock_function(self, obj, func):
        """
        Replaces a bounded function in `self.breaker` by another.
        """
        setattr(obj, func.__name__, MethodType(func, self.breaker))

    def test_fail_thread_safety(self):
        """CircuitBreaker: it should compute a failed call atomically to
        avoid race conditions.
        """
        # Create a specific exception to avoid masking other errors
        class SpecificException(Exception):
            pass

        @self.breaker
        def err(): raise SpecificException()

        def trigger_error():
            for n in range(500):
                try: err()
                except SpecificException: pass

        def _inc_counter(self):
            sleep(0.00005)
            self._state_storage.increment_counter()

        self._mock_function(self.breaker, _inc_counter)
        self._start_threads(trigger_error, 3)
        self.assertEqual(1500, self.breaker.fail_counter)

    def test_success_thread_safety(self):
        """CircuitBreaker: it should compute a successful call atomically
        to avoid race conditions.
        """
        @self.breaker
        def suc(): return True

        def trigger_success():
            for n in range(500):
                suc()

        class SuccessListener(CircuitBreakerListener):
            def success(self, cb):
                c = 0
                if hasattr(cb, '_success_counter'):
                    c = cb._success_counter
                sleep(0.00005)
                cb._success_counter = c + 1

        self.breaker.add_listener(SuccessListener())
        self._start_threads(trigger_success, 3)
        self.assertEqual(1500, self.breaker._success_counter)

    def test_half_open_thread_safety(self):
        """CircuitBreaker: it should allow only one trial call when the
        circuit is half-open.
        """
        self.breaker = CircuitBreaker(fail_max=1, reset_timeout=0.01)

        self.breaker.open()
        sleep(0.01)

        @self.breaker
        def err(): raise Exception()

        def trigger_failure():
            try: err()
            except: pass

        class StateListener(CircuitBreakerListener):
            def __init__(self):
                self._count = 0

            def before_call(self, cb, fun, *args, **kwargs):
                sleep(0.00005)

            def state_change(self, cb, old_state, new_state):
                if new_state.name == 'half-open':
                    self._count += 1

        state_listener = StateListener()
        self.breaker.add_listener(state_listener)

        self._start_threads(trigger_failure, 5)
        self.assertEqual(1, state_listener._count)


    def test_fail_max_thread_safety(self):
        """CircuitBreaker: it should not allow more failed calls than 'fail_max'
        setting. Note that with Redis, where we have separate systems
        incrementing the counter, we can get concurrent updates such that the
        counter is greater than the 'fail_max' by the number of systems. To
        prevent this, we'd need to take out a lock amongst all systems before
        trying the call.
        """
        @self.breaker
        def err(): raise Exception()

        def trigger_error():
            for i in range(2000):
                try: err()
                except: pass

        class SleepListener(CircuitBreakerListener):
            def before_call(self, cb, func, *args, **kwargs):
                sleep(0.00005)

        self.breaker.add_listener(SleepListener())
        num_threads = 3
        self._start_threads(trigger_error, num_threads)
        self.assertTrue(self.breaker.fail_counter < self.breaker.fail_max + num_threads)


if __name__ == "__main__":
    unittest.main()




# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Processors and helpers specific to the :mod:`logging` module from the `Python
standard library <https://docs.python.org/>`_.

See also :doc:`structlog's standard library support <standard-library>`.
"""

from __future__ import absolute_import, division, print_function

import logging

from six import PY3

from structlog._base import BoundLoggerBase
from structlog._frames import _find_first_app_frame_and_name, _format_stack
from structlog.exceptions import DropEvent


class _FixedFindCallerLogger(logging.Logger):
    """
    Change the behavior of `logging.Logger.findCaller` to cope with
    ``structlog``'s extra frames.
    """

    def findCaller(self, stack_info=False, stacklevel=1):
        """
        Finds the first caller frame outside of structlog so that the caller
        info is populated for wrapping stdlib.

        This logger gets set as the default one when using LoggerFactory.
        """
        f, name = _find_first_app_frame_and_name(["logging"])
        if PY3:
            if stack_info:
                sinfo = _format_stack(f)
            else:
                sinfo = None
            return f.f_code.co_filename, f.f_lineno, f.f_code.co_name, sinfo
        else:
            return f.f_code.co_filename, f.f_lineno, f.f_code.co_name


class BoundLogger(BoundLoggerBase):
    """
    Python Standard Library version of `structlog.BoundLogger`.

    Works exactly like the generic one except that it takes advantage of
    knowing the logging methods in advance.

    Use it like::

        structlog.configure(
            wrapper_class=structlog.stdlib.BoundLogger,
        )

    It also contains a bunch of properties that pass-through to the wrapped
    `logging.Logger` which should make it work as a drop-in replacement.
    """

    def debug(self, event=None, *args, **kw):
        """
        Process event and call `logging.Logger.debug` with the result.
        """
        return self._proxy_to_logger("debug", event, *args, **kw)

    def info(self, event=None, *args, **kw):
        """
        Process event and call `logging.Logger.info` with the result.
        """
        return self._proxy_to_logger("info", event, *args, **kw)

    def warning(self, event=None, *args, **kw):
        """
        Process event and call `logging.Logger.warning` with the result.
        """
        return self._proxy_to_logger("warning", event, *args, **kw)

    warn = warning

    def error(self, event=None, *args, **kw):
        """
        Process event and call `logging.Logger.error` with the result.
        """
        return self._proxy_to_logger("error", event, *args, **kw)

    def critical(self, event=None, *args, **kw):
        """
        Process event and call `logging.Logger.critical` with the result.
        """
        return self._proxy_to_logger("critical", event, *args, **kw)

    def exception(self, event=None, *args, **kw):
        """
        Process event and call `logging.Logger.error` with the result,
        after setting ``exc_info`` to `True`.
        """
        kw.setdefault("exc_info", True)
        return self.error(event, *args, **kw)

    def log(self, level, event, *args, **kw):
        """
        Process *event* and call the appropriate logging method depending on
        *level*.
        """
        return self._proxy_to_logger(_LEVEL_TO_NAME[level], event, *args, **kw)

    fatal = critical

    def _proxy_to_logger(self, method_name, event, *event_args, **event_kw):
        """
        Propagate a method call to the wrapped logger.

        This is the same as the superclass implementation, except that
        it also preserves positional arguments in the ``event_dict`` so
        that the stdblib's support for format strings can be used.
        """
        if event_args:
            event_kw["positional_args"] = event_args
        return super(BoundLogger, self)._proxy_to_logger(
            method_name, event=event, **event_kw
        )

    #
    # Pass-through attributes and methods to mimick the stdlib's logger
    # interface.
    #

    @property
    def name(self):
        """
        Returns :attr:`logging.Logger.name`
        """
        return self._logger.name

    @property
    def level(self):
        """
        Returns :attr:`logging.Logger.level`
        """
        return self._logger.level

    @property
    def parent(self):
        """
        Returns :attr:`logging.Logger.parent`
        """
        return self._logger.parent

    @property
    def propagate(self):
        """
        Returns :attr:`logging.Logger.propagate`
        """
        return self._logger.propagate

    @property
    def handlers(self):
        """
        Returns :attr:`logging.Logger.handlers`
        """
        return self._logger.handlers

    @property
    def disabled(self):
        """
        Returns :attr:`logging.Logger.disabled`
        """
        return self._logger.disabled

    def setLevel(self, level):
        """
        Calls :meth:`logging.Logger.setLevel` with unmodified arguments.
        """
        self._logger.setLevel(level)

    def findCaller(self, stack_info=False):
        """
        Calls :meth:`logging.Logger.findCaller` with unmodified arguments.
        """
        return self._logger.findCaller(stack_info=stack_info)

    def makeRecord(
        self, name, level, fn, lno, msg, args, exc_info, func=None, extra=None
    ):
        """
        Calls :meth:`logging.Logger.makeRecord` with unmodified arguments.
        """
        return self._logger.makeRecord(
            name, level, fn, lno, msg, args, exc_info, func=func, extra=extra
        )

    def handle(self, record):
        """
        Calls :meth:`logging.Logger.handle` with unmodified arguments.
        """
        self._logger.handle(record)

    def addHandler(self, hdlr):
        """
        Calls :meth:`logging.Logger.addHandler` with unmodified arguments.
        """
        self._logger.addHandler(hdlr)

    def removeHandler(self, hdlr):
        """
        Calls :meth:`logging.Logger.removeHandler` with unmodified arguments.
        """
        self._logger.removeHandler(hdlr)

    def hasHandlers(self):
        """
        Calls :meth:`logging.Logger.hasHandlers` with unmodified arguments.

        Exists only in Python 3.
        """
        return self._logger.hasHandlers()

    def callHandlers(self, record):
        """
        Calls :meth:`logging.Logger.callHandlers` with unmodified arguments.
        """
        self._logger.callHandlers(record)

    def getEffectiveLevel(self):
        """
        Calls :meth:`logging.Logger.getEffectiveLevel` with unmodified
        arguments.
        """
        return self._logger.getEffectiveLevel()

    def isEnabledFor(self, level):
        """
        Calls :meth:`logging.Logger.isEnabledFor` with unmodified arguments.
        """
        return self._logger.isEnabledFor(level)

    def getChild(self, suffix):
        """
        Calls :meth:`logging.Logger.getChild` with unmodified arguments.
        """
        return self._logger.getChild(suffix)


class LoggerFactory(object):
    """
    Build a standard library logger when an *instance* is called.

    Sets a custom logger using :func:`logging.setLoggerClass` so variables in
    log format are expanded properly.

    >>> from structlog import configure
    >>> from structlog.stdlib import LoggerFactory
    >>> configure(logger_factory=LoggerFactory())

    :param ignore_frame_names: When guessing the name of a logger, skip frames
        whose names *start* with one of these.  For example, in pyramid
        applications you'll want to set it to
        ``["venusian", "pyramid.config"]``.
    :type ignore_frame_names: ``list`` of ``str``
    """

    def __init__(self, ignore_frame_names=None):
        self._ignore = ignore_frame_names
        logging.setLoggerClass(_FixedFindCallerLogger)

    def __call__(self, *args):
        """
        Deduce the caller's module name and create a stdlib logger.

        If an optional argument is passed, it will be used as the logger name
        instead of guesswork.  This optional argument would be passed from the
        :func:`structlog.get_logger` call.  For example
        ``structlog.get_logger("foo")`` would cause this method to be called
        with ``"foo"`` as its first positional argument.

        :rtype: logging.Logger

        .. versionchanged:: 0.4.0
            Added support for optional positional arguments.  Using the first
            one for naming the constructed logger.
        """
        if args:
            return logging.getLogger(args[0])

        # We skip all frames that originate from within structlog or one of the
        # configured names.
        _, name = _find_first_app_frame_and_name(self._ignore)
        return logging.getLogger(name)


class PositionalArgumentsFormatter(object):
    """
    Apply stdlib-like string formatting to the ``event`` key.

    If the ``positional_args`` key in the event dict is set, it must
    contain a tuple that is used for formatting (using the ``%s`` string
    formatting operator) of the value from the ``event`` key.  This works
    in the same way as the stdlib handles arguments to the various log
    methods: if the tuple contains only a single `dict` argument it is
    used for keyword placeholders in the ``event`` string, otherwise it
    will be used for positional placeholders.

    ``positional_args`` is populated by `structlog.stdlib.BoundLogger` or
    can be set manually.

    The *remove_positional_args* flag can be set to `False` to keep the
    ``positional_args`` key in the event dict; by default it will be
    removed from the event dict after formatting a message.
    """

    def __init__(self, remove_positional_args=True):
        self.remove_positional_args = remove_positional_args

    def __call__(self, _, __, event_dict):
        args = event_dict.get("positional_args")

        # Mimick the formatting behaviour of the stdlib's logging
        # module, which accepts both positional arguments and a single
        # dict argument. The "single dict" check is the same one as the
        # stdlib's logging module performs in LogRecord.__init__().
        if args:
            if len(args) == 1 and isinstance(args[0], dict) and args[0]:
                args = args[0]
            event_dict["event"] = event_dict["event"] % args
        if self.remove_positional_args and args is not None:
            del event_dict["positional_args"]
        return event_dict


# Adapted from the stdlib
CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0

_NAME_TO_LEVEL = {
    "critical": CRITICAL,
    "exception": ERROR,
    "error": ERROR,
    "warn": WARNING,
    "warning": WARNING,
    "info": INFO,
    "debug": DEBUG,
    "notset": NOTSET,
}

_LEVEL_TO_NAME = dict(
    (v, k)
    for k, v in _NAME_TO_LEVEL.items()
    if k not in ("warn", "exception", "notset")
)


def filter_by_level(logger, name, event_dict):
    """
    Check whether logging is configured to accept messages from this log level.

    Should be the first processor if stdlib's filtering by level is used so
    possibly expensive processors like exception formatters are avoided in the
    first place.

    >>> import logging
    >>> from structlog.stdlib import filter_by_level
    >>> logging.basicConfig(level=logging.WARN)
    >>> logger = logging.getLogger()
    >>> filter_by_level(logger, 'warn', {})
    {}
    >>> filter_by_level(logger, 'debug', {})
    Traceback (most recent call last):
    ...
    DropEvent
    """
    if logger.isEnabledFor(_NAME_TO_LEVEL[name]):
        return event_dict
    else:
        raise DropEvent


def add_log_level(logger, method_name, event_dict):
    """
    Add the log level to the event dict.
    """
    if method_name == "warn":
        # The stdlib has an alias
        method_name = "warning"

    event_dict["level"] = method_name
    return event_dict


def add_log_level_number(logger, method_name, event_dict):
    """
    Add the log level number to the event dict.

    Log level numbers map to the log level names. The Python stdlib uses them
    for filtering logic. This adds the same numbers so users can leverage
    similar filtering. Compare::

       level in ("warning", "error", "critical")
       level_number >= 30

    The mapping of names to numbers is in ``structlog.stdlib._NAME_TO_LEVEL``.

    .. versionadded:: 18.2.0
    """
    event_dict["level_number"] = _NAME_TO_LEVEL[method_name]
    return event_dict


def add_logger_name(logger, method_name, event_dict):
    """
    Add the logger name to the event dict.
    """
    record = event_dict.get("_record")
    if record is None:
        event_dict["logger"] = logger.name
    else:
        event_dict["logger"] = record.name
    return event_dict


def render_to_log_kwargs(wrapped_logger, method_name, event_dict):
    """
    Render ``event_dict`` into keyword arguments for `logging.log`.

    The ``event`` field is translated into ``msg`` and the rest of the
    *event_dict* is added as ``extra``.

    This allows you to defer formatting to `logging`.

    .. versionadded:: 17.1.0
    """
    return {"msg": event_dict.pop("event"), "extra": event_dict}


class ProcessorFormatter(logging.Formatter):
    r"""
    Call ``structlog`` processors on :`logging.LogRecord`\ s.

    This `logging.Formatter` allows to configure :mod:`logging` to call
    *processor* on ``structlog``-borne log entries (origin is determined solely
    on the fact whether the ``msg`` field on the `logging.LogRecord` is
    a dict or not).

    This allows for two interesting use cases:

    #. You can format non-``structlog`` log entries.
    #. You can multiplex log records into multiple `logging.Handler`\ s.

    Please refer to :doc:`standard-library` for examples.

    :param callable processor: A ``structlog`` processor.
    :param foreign_pre_chain:
        If not `None`, it is used as an iterable of processors that is applied
        to non-``structlog`` log entries before *processor*.  If `None`,
        formatting is left to :mod:`logging`. (default: `None`)
    :param bool keep_exc_info: ``exc_info`` on `logging.LogRecord`\ s is
        added to the ``event_dict`` and removed afterwards. Set this to
        ``True`` to keep it on the `logging.LogRecord`. (default: False)
    :param bool keep_stack_info: Same as *keep_exc_info* except for Python 3's
        ``stack_info``. (default: False)
    :param logger: Logger which we want to push through the ``structlog``
        processor chain. This parameter is necessary for some of the
        processors like `filter_by_level`. (default: None)
    :param bool pass_foreign_args: If True, pass a foreign log record's
        ``args`` attribute to the ``event_dict`` under ``positional_args`` key.
        (default: False)
    :rtype: str

    .. versionadded:: 17.1.0
    .. versionadded:: 17.2.0 *keep_exc_info* and *keep_stack_info*
    .. versionadded:: 19.2.0 *logger*
    .. versionadded:: 19.2.0 *pass_foreign_args*
    """

    def __init__(
        self,
        processor,
        foreign_pre_chain=None,
        keep_exc_info=False,
        keep_stack_info=False,
        logger=None,
        pass_foreign_args=False,
        *args,
        **kwargs
    ):
        fmt = kwargs.pop("fmt", "%(message)s")
        super(ProcessorFormatter, self).__init__(*args, fmt=fmt, **kwargs)
        self.processor = processor
        self.foreign_pre_chain = foreign_pre_chain
        self.keep_exc_info = keep_exc_info
        # The and clause saves us checking for PY3 in the formatter.
        self.keep_stack_info = keep_stack_info and PY3
        self.logger = logger
        self.pass_foreign_args = pass_foreign_args

    def format(self, record):
        """
        Extract ``structlog``'s `event_dict` from ``record.msg`` and format it.
        """
        # Make a shallow copy of the record to let other handlers/formatters
        # process the original one
        record = logging.makeLogRecord(record.__dict__)
        try:
            # Both attached by wrap_for_formatter
            logger = self.logger if self.logger is not None else record._logger
            meth_name = record._name

            # We need to copy because it's possible that the same record gets
            # processed by multiple logging formatters.  LogRecord.getMessage
            # would transform our dict into a str.
            ed = record.msg.copy()
        except AttributeError:
            logger = self.logger
            meth_name = record.levelname.lower()
            ed = {"event": record.getMessage(), "_record": record}

            if self.pass_foreign_args:
                ed["positional_args"] = record.args

            record.args = ()

            # Add stack-related attributes to event_dict and unset them
            # on the record copy so that the base implementation wouldn't
            # append stacktraces to the output.
            if record.exc_info:
                ed["exc_info"] = record.exc_info
            if PY3 and record.stack_info:
                ed["stack_info"] = record.stack_info

            if not self.keep_exc_info:
                record.exc_text = None
                record.exc_info = None
            if not self.keep_stack_info:
                record.stack_info = None

            # Non-structlog allows to run through a chain to prepare it for the
            # final processor (e.g. adding timestamps and log levels).
            for proc in self.foreign_pre_chain or ():
                ed = proc(logger, meth_name, ed)

            del ed["_record"]

        record.msg = self.processor(logger, meth_name, ed)
        return super(ProcessorFormatter, self).format(record)

    @staticmethod
    def wrap_for_formatter(logger, name, event_dict):
        """
        Wrap *logger*, *name*, and *event_dict*.

        The result is later unpacked by `ProcessorFormatter` when
        formatting log entries.

        Use this static method as the renderer (i.e. final processor) if you
        want to use `ProcessorFormatter` in your `logging` configuration.
        """
        return (event_dict,), {"extra": {"_logger": logger, "_name": name}}



import os
import sys

from pip._vendor.six.moves.urllib import parse as urllib_parse
from pip._vendor.six.moves.urllib import request as urllib_request

from pip._internal.utils.typing import MYPY_CHECK_RUNNING

if MYPY_CHECK_RUNNING:
    from typing import Optional, Text, Union


def get_url_scheme(url):
    # type: (Union[str, Text]) -> Optional[Text]
    if ':' not in url:
        return None
    return url.split(':', 1)[0].lower()


def path_to_url(path):
    # type: (Union[str, Text]) -> str
    """
    Convert a path to a file: URL.  The path will be made absolute and have
    quoted path parts.
    """
    path = os.path.normpath(os.path.abspath(path))
    url = urllib_parse.urljoin('file:', urllib_request.pathname2url(path))
    return url


def url_to_path(url):
    # type: (str) -> str
    """
    Convert a file: URL to a path.
    """
    assert url.startswith('file:'), (
        "You can only turn file: urls into filenames (not {url!r})"
        .format(**locals()))

    _, netloc, path, _, _ = urllib_parse.urlsplit(url)

    if not netloc or netloc == 'localhost':
        # According to RFC 8089, same as empty authority.
        netloc = ''
    elif sys.platform == 'win32':
        # If we have a UNC path, prepend UNC share notation.
        netloc = '\\\\' + netloc
    else:
        raise ValueError(
            'non-local file URIs are not supported on this platform: {url!r}'
            .format(**locals())
        )

    path = urllib_request.url2pathname(netloc + path)
    return path



"""All the custom exceptions used within gtransport are declared here"""


class StrictModeError(ValueError):
    """
    To be raised when the exceptional condition which could have
    been ignored normally, but being raised as the application
    is running in `strict` mode.

    Such errors will normally be due to invalid or missing input.
    Hence, this class extends `ValueError`
    """
    pass



from middlewares.django_middlewares import DjangoMiddleware
from glogger import (
    configure,
    new_logger,
    new_bounded_logger,
    get_logger,
    getLogger,
    bind,
    get_trace_id
)

__all__ = [
    "DjangoMiddleware",
    "configure",
    "new_logger",
    "new_bounded_logger",
    "get_logger",
    "getLogger",
    "bind",
    "get_trace_id"
]



# -*- coding: utf-8 -*-

#   __
#  /__)  _  _     _   _ _/   _
# / (   (- (/ (/ (- _)  /  _)
#          /

"""
Requests HTTP Library
~~~~~~~~~~~~~~~~~~~~~

Requests is an HTTP library, written in Python, for human beings.
Basic GET usage:

   >>> import requests
   >>> r = requests.get('https://www.python.org')
   >>> r.status_code
   200
   >>> b'Python is a programming language' in r.content
   True

... or POST:

   >>> payload = dict(key1='value1', key2='value2')
   >>> r = requests.post('https://httpbin.org/post', data=payload)
   >>> print(r.text)
   {
     ...
     "form": {
       "key1": "value1",
       "key2": "value2"
     },
     ...
   }

The other HTTP methods are supported - see `requests.api`. Full documentation
is at <https://requests.readthedocs.io>.

:copyright: (c) 2017 by Kenneth Reitz.
:license: Apache 2.0, see LICENSE for more details.
"""

from pip._vendor import urllib3
from pip._vendor import chardet
import warnings
from .exceptions import RequestsDependencyWarning


def check_compatibility(urllib3_version, chardet_version):
    urllib3_version = urllib3_version.split('.')
    assert urllib3_version != ['dev']  # Verify urllib3 isn't installed from git.

    # Sometimes, urllib3 only reports its version as 16.1.
    if len(urllib3_version) == 2:
        urllib3_version.append('0')

    # Check urllib3 for compatibility.
    major, minor, patch = urllib3_version  # noqa: F811
    major, minor, patch = int(major), int(minor), int(patch)
    # urllib3 >= 1.21.1, <= 1.26
    assert major == 1
    assert minor >= 21
    assert minor <= 26

    # Check chardet for compatibility.
    major, minor, patch = chardet_version.split('.')[:3]
    major, minor, patch = int(major), int(minor), int(patch)
    # chardet >= 3.0.2, < 3.1.0
    assert major == 3
    assert minor < 1
    assert patch >= 2


def _check_cryptography(cryptography_version):
    # cryptography < 1.3.4
    try:
        cryptography_version = list(map(int, cryptography_version.split('.')))
    except ValueError:
        return

    if cryptography_version < [1, 3, 4]:
        warning = 'Old version of cryptography ({}) may cause slowdown.'.format(cryptography_version)
        warnings.warn(warning, RequestsDependencyWarning)

# Check imported dependencies for compatibility.
try:
    check_compatibility(urllib3.__version__, chardet.__version__)
except (AssertionError, ValueError):
    warnings.warn("urllib3 ({}) or chardet ({}) doesn't match a supported "
                  "version!".format(urllib3.__version__, chardet.__version__),
                  RequestsDependencyWarning)

# Attempt to enable urllib3's fallback for SNI support
# if the standard library doesn't support SNI or the
# 'ssl' library isn't available.
try:
    # Note: This logic prevents upgrading cryptography on Windows, if imported
    #       as part of pip.
    from pip._internal.utils.compat import WINDOWS
    if not WINDOWS:
        raise ImportError("pip internals: don't import cryptography on Windows")
    try:
        import ssl
    except ImportError:
        ssl = None

    if not getattr(ssl, "HAS_SNI", False):
        from pip._vendor.urllib3.contrib import pyopenssl
        pyopenssl.inject_into_urllib3()

        # Check cryptography version
        from cryptography import __version__ as cryptography_version
        _check_cryptography(cryptography_version)
except ImportError:
    pass

# urllib3's DependencyWarnings should be silenced.
from pip._vendor.urllib3.exceptions import DependencyWarning
warnings.simplefilter('ignore', DependencyWarning)

from .__version__ import __title__, __description__, __url__, __version__
from .__version__ import __build__, __author__, __author_email__, __license__
from .__version__ import __copyright__, __cake__

from . import utils
from . import packages
from .models import Request, Response, PreparedRequest
from .api import request, get, head, post, patch, put, delete, options
from .sessions import session, Session
from .status_codes import codes
from .exceptions import (
    RequestException, Timeout, URLRequired,
    TooManyRedirects, HTTPError, ConnectionError,
    FileModeWarning, ConnectTimeout, ReadTimeout
)

# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())

# FileModeWarnings go off per the default.
warnings.simplefilter('default', FileModeWarning, append=True)



# -*- coding: utf-8 -*-
"""
    pygments.lexers._vim_builtins
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This file is autogenerated by scripts/get_vimkw.py

    :copyright: Copyright 2006-2019 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

# Split up in multiple functions so it's importable by jython, which has a
# per-method size limit.

def _getauto():
    var = (
        ('BufAdd','BufAdd'),
        ('BufCreate','BufCreate'),
        ('BufDelete','BufDelete'),
        ('BufEnter','BufEnter'),
        ('BufFilePost','BufFilePost'),
        ('BufFilePre','BufFilePre'),
        ('BufHidden','BufHidden'),
        ('BufLeave','BufLeave'),
        ('BufNew','BufNew'),
        ('BufNewFile','BufNewFile'),
        ('BufRead','BufRead'),
        ('BufReadCmd','BufReadCmd'),
        ('BufReadPost','BufReadPost'),
        ('BufReadPre','BufReadPre'),
        ('BufUnload','BufUnload'),
        ('BufWinEnter','BufWinEnter'),
        ('BufWinLeave','BufWinLeave'),
        ('BufWipeout','BufWipeout'),
        ('BufWrite','BufWrite'),
        ('BufWriteCmd','BufWriteCmd'),
        ('BufWritePost','BufWritePost'),
        ('BufWritePre','BufWritePre'),
        ('Cmd','Cmd'),
        ('CmdwinEnter','CmdwinEnter'),
        ('CmdwinLeave','CmdwinLeave'),
        ('ColorScheme','ColorScheme'),
        ('CompleteDone','CompleteDone'),
        ('CursorHold','CursorHold'),
        ('CursorHoldI','CursorHoldI'),
        ('CursorMoved','CursorMoved'),
        ('CursorMovedI','CursorMovedI'),
        ('EncodingChanged','EncodingChanged'),
        ('FileAppendCmd','FileAppendCmd'),
        ('FileAppendPost','FileAppendPost'),
        ('FileAppendPre','FileAppendPre'),
        ('FileChangedRO','FileChangedRO'),
        ('FileChangedShell','FileChangedShell'),
        ('FileChangedShellPost','FileChangedShellPost'),
        ('FileEncoding','FileEncoding'),
        ('FileReadCmd','FileReadCmd'),
        ('FileReadPost','FileReadPost'),
        ('FileReadPre','FileReadPre'),
        ('FileType','FileType'),
        ('FileWriteCmd','FileWriteCmd'),
        ('FileWritePost','FileWritePost'),
        ('FileWritePre','FileWritePre'),
        ('FilterReadPost','FilterReadPost'),
        ('FilterReadPre','FilterReadPre'),
        ('FilterWritePost','FilterWritePost'),
        ('FilterWritePre','FilterWritePre'),
        ('FocusGained','FocusGained'),
        ('FocusLost','FocusLost'),
        ('FuncUndefined','FuncUndefined'),
        ('GUIEnter','GUIEnter'),
        ('GUIFailed','GUIFailed'),
        ('InsertChange','InsertChange'),
        ('InsertCharPre','InsertCharPre'),
        ('InsertEnter','InsertEnter'),
        ('InsertLeave','InsertLeave'),
        ('MenuPopup','MenuPopup'),
        ('QuickFixCmdPost','QuickFixCmdPost'),
        ('QuickFixCmdPre','QuickFixCmdPre'),
        ('QuitPre','QuitPre'),
        ('RemoteReply','RemoteReply'),
        ('SessionLoadPost','SessionLoadPost'),
        ('ShellCmdPost','ShellCmdPost'),
        ('ShellFilterPost','ShellFilterPost'),
        ('SourceCmd','SourceCmd'),
        ('SourcePre','SourcePre'),
        ('SpellFileMissing','SpellFileMissing'),
        ('StdinReadPost','StdinReadPost'),
        ('StdinReadPre','StdinReadPre'),
        ('SwapExists','SwapExists'),
        ('Syntax','Syntax'),
        ('TabEnter','TabEnter'),
        ('TabLeave','TabLeave'),
        ('TermChanged','TermChanged'),
        ('TermResponse','TermResponse'),
        ('TextChanged','TextChanged'),
        ('TextChangedI','TextChangedI'),
        ('User','User'),
        ('UserGettingBored','UserGettingBored'),
        ('VimEnter','VimEnter'),
        ('VimLeave','VimLeave'),
        ('VimLeavePre','VimLeavePre'),
        ('VimResized','VimResized'),
        ('WinEnter','WinEnter'),
        ('WinLeave','WinLeave'),
        ('event','event'),
    )
    return var
auto = _getauto()

def _getcommand():
    var = (
        ('a','a'),
        ('ab','ab'),
        ('abc','abclear'),
        ('abo','aboveleft'),
        ('al','all'),
        ('ar','ar'),
        ('ar','args'),
        ('arga','argadd'),
        ('argd','argdelete'),
        ('argdo','argdo'),
        ('arge','argedit'),
        ('argg','argglobal'),
        ('argl','arglocal'),
        ('argu','argument'),
        ('as','ascii'),
        ('au','au'),
        ('b','buffer'),
        ('bN','bNext'),
        ('ba','ball'),
        ('bad','badd'),
        ('bd','bdelete'),
        ('bel','belowright'),
        ('bf','bfirst'),
        ('bl','blast'),
        ('bm','bmodified'),
        ('bn','bnext'),
        ('bo','botright'),
        ('bp','bprevious'),
        ('br','br'),
        ('br','brewind'),
        ('brea','break'),
        ('breaka','breakadd'),
        ('breakd','breakdel'),
        ('breakl','breaklist'),
        ('bro','browse'),
        ('bu','bu'),
        ('buf','buf'),
        ('bufdo','bufdo'),
        ('buffers','buffers'),
        ('bun','bunload'),
        ('bw','bwipeout'),
        ('c','c'),
        ('c','change'),
        ('cN','cN'),
        ('cN','cNext'),
        ('cNf','cNf'),
        ('cNf','cNfile'),
        ('cabc','cabclear'),
        ('cad','cad'),
        ('cad','caddexpr'),
        ('caddb','caddbuffer'),
        ('caddf','caddfile'),
        ('cal','call'),
        ('cat','catch'),
        ('cb','cbuffer'),
        ('cc','cc'),
        ('ccl','cclose'),
        ('cd','cd'),
        ('ce','center'),
        ('cex','cexpr'),
        ('cf','cfile'),
        ('cfir','cfirst'),
        ('cg','cgetfile'),
        ('cgetb','cgetbuffer'),
        ('cgete','cgetexpr'),
        ('changes','changes'),
        ('chd','chdir'),
        ('che','checkpath'),
        ('checkt','checktime'),
        ('cl','cl'),
        ('cl','clist'),
        ('cla','clast'),
        ('clo','close'),
        ('cmapc','cmapclear'),
        ('cn','cn'),
        ('cn','cnext'),
        ('cnew','cnewer'),
        ('cnf','cnf'),
        ('cnf','cnfile'),
        ('co','copy'),
        ('col','colder'),
        ('colo','colorscheme'),
        ('com','com'),
        ('comc','comclear'),
        ('comp','compiler'),
        ('con','con'),
        ('con','continue'),
        ('conf','confirm'),
        ('cope','copen'),
        ('cp','cprevious'),
        ('cpf','cpfile'),
        ('cq','cquit'),
        ('cr','crewind'),
        ('cs','cs'),
        ('cscope','cscope'),
        ('cstag','cstag'),
        ('cuna','cunabbrev'),
        ('cw','cwindow'),
        ('d','d'),
        ('d','delete'),
        ('de','de'),
        ('debug','debug'),
        ('debugg','debuggreedy'),
        ('del','del'),
        ('delc','delcommand'),
        ('delel','delel'),
        ('delep','delep'),
        ('deletel','deletel'),
        ('deletep','deletep'),
        ('deletl','deletl'),
        ('deletp','deletp'),
        ('delf','delf'),
        ('delf','delfunction'),
        ('dell','dell'),
        ('delm','delmarks'),
        ('delp','delp'),
        ('dep','dep'),
        ('di','di'),
        ('di','display'),
        ('diffg','diffget'),
        ('diffo','diffoff'),
        ('diffp','diffpatch'),
        ('diffpu','diffput'),
        ('diffs','diffsplit'),
        ('difft','diffthis'),
        ('diffu','diffupdate'),
        ('dig','dig'),
        ('dig','digraphs'),
        ('dir','dir'),
        ('dj','djump'),
        ('dl','dl'),
        ('dli','dlist'),
        ('do','do'),
        ('doau','doau'),
        ('dp','dp'),
        ('dr','drop'),
        ('ds','dsearch'),
        ('dsp','dsplit'),
        ('e','e'),
        ('e','edit'),
        ('ea','ea'),
        ('earlier','earlier'),
        ('ec','ec'),
        ('echoe','echoerr'),
        ('echom','echomsg'),
        ('echon','echon'),
        ('el','else'),
        ('elsei','elseif'),
        ('em','emenu'),
        ('en','en'),
        ('en','endif'),
        ('endf','endf'),
        ('endf','endfunction'),
        ('endfo','endfor'),
        ('endfun','endfun'),
        ('endt','endtry'),
        ('endw','endwhile'),
        ('ene','enew'),
        ('ex','ex'),
        ('exi','exit'),
        ('exu','exusage'),
        ('f','f'),
        ('f','file'),
        ('files','files'),
        ('filet','filet'),
        ('filetype','filetype'),
        ('fin','fin'),
        ('fin','find'),
        ('fina','finally'),
        ('fini','finish'),
        ('fir','first'),
        ('fix','fixdel'),
        ('fo','fold'),
        ('foldc','foldclose'),
        ('foldd','folddoopen'),
        ('folddoc','folddoclosed'),
        ('foldo','foldopen'),
        ('for','for'),
        ('fu','fu'),
        ('fu','function'),
        ('fun','fun'),
        ('g','g'),
        ('go','goto'),
        ('gr','grep'),
        ('grepa','grepadd'),
        ('gui','gui'),
        ('gvim','gvim'),
        ('h','h'),
        ('h','help'),
        ('ha','hardcopy'),
        ('helpf','helpfind'),
        ('helpg','helpgrep'),
        ('helpt','helptags'),
        ('hi','hi'),
        ('hid','hide'),
        ('his','history'),
        ('i','i'),
        ('ia','ia'),
        ('iabc','iabclear'),
        ('if','if'),
        ('ij','ijump'),
        ('il','ilist'),
        ('imapc','imapclear'),
        ('in','in'),
        ('intro','intro'),
        ('is','isearch'),
        ('isp','isplit'),
        ('iuna','iunabbrev'),
        ('j','join'),
        ('ju','jumps'),
        ('k','k'),
        ('kee','keepmarks'),
        ('keepa','keepa'),
        ('keepalt','keepalt'),
        ('keepj','keepjumps'),
        ('keepp','keeppatterns'),
        ('l','l'),
        ('l','list'),
        ('lN','lN'),
        ('lN','lNext'),
        ('lNf','lNf'),
        ('lNf','lNfile'),
        ('la','la'),
        ('la','last'),
        ('lad','lad'),
        ('lad','laddexpr'),
        ('laddb','laddbuffer'),
        ('laddf','laddfile'),
        ('lan','lan'),
        ('lan','language'),
        ('lat','lat'),
        ('later','later'),
        ('lb','lbuffer'),
        ('lc','lcd'),
        ('lch','lchdir'),
        ('lcl','lclose'),
        ('lcs','lcs'),
        ('lcscope','lcscope'),
        ('le','left'),
        ('lefta','leftabove'),
        ('lex','lexpr'),
        ('lf','lfile'),
        ('lfir','lfirst'),
        ('lg','lgetfile'),
        ('lgetb','lgetbuffer'),
        ('lgete','lgetexpr'),
        ('lgr','lgrep'),
        ('lgrepa','lgrepadd'),
        ('lh','lhelpgrep'),
        ('ll','ll'),
        ('lla','llast'),
        ('lli','llist'),
        ('lmak','lmake'),
        ('lmapc','lmapclear'),
        ('lne','lne'),
        ('lne','lnext'),
        ('lnew','lnewer'),
        ('lnf','lnf'),
        ('lnf','lnfile'),
        ('lo','lo'),
        ('lo','loadview'),
        ('loadk','loadk'),
        ('loadkeymap','loadkeymap'),
        ('loc','lockmarks'),
        ('lockv','lockvar'),
        ('lol','lolder'),
        ('lop','lopen'),
        ('lp','lprevious'),
        ('lpf','lpfile'),
        ('lr','lrewind'),
        ('ls','ls'),
        ('lt','ltag'),
        ('lua','lua'),
        ('luado','luado'),
        ('luafile','luafile'),
        ('lv','lvimgrep'),
        ('lvimgrepa','lvimgrepadd'),
        ('lw','lwindow'),
        ('m','move'),
        ('ma','ma'),
        ('ma','mark'),
        ('mak','make'),
        ('marks','marks'),
        ('mat','match'),
        ('menut','menut'),
        ('menut','menutranslate'),
        ('mes','mes'),
        ('messages','messages'),
        ('mk','mk'),
        ('mk','mkexrc'),
        ('mks','mksession'),
        ('mksp','mkspell'),
        ('mkv','mkv'),
        ('mkv','mkvimrc'),
        ('mkvie','mkview'),
        ('mo','mo'),
        ('mod','mode'),
        ('mz','mz'),
        ('mz','mzscheme'),
        ('mzf','mzfile'),
        ('n','n'),
        ('n','next'),
        ('nb','nbkey'),
        ('nbc','nbclose'),
        ('nbs','nbstart'),
        ('ne','ne'),
        ('new','new'),
        ('nmapc','nmapclear'),
        ('noa','noa'),
        ('noautocmd','noautocmd'),
        ('noh','nohlsearch'),
        ('nu','number'),
        ('o','o'),
        ('o','open'),
        ('ol','oldfiles'),
        ('omapc','omapclear'),
        ('on','only'),
        ('opt','options'),
        ('ownsyntax','ownsyntax'),
        ('p','p'),
        ('p','print'),
        ('pc','pclose'),
        ('pe','pe'),
        ('pe','perl'),
        ('ped','pedit'),
        ('perld','perldo'),
        ('po','pop'),
        ('popu','popu'),
        ('popu','popup'),
        ('pp','ppop'),
        ('pr','pr'),
        ('pre','preserve'),
        ('prev','previous'),
        ('pro','pro'),
        ('prof','profile'),
        ('profd','profdel'),
        ('promptf','promptfind'),
        ('promptr','promptrepl'),
        ('ps','psearch'),
        ('ptN','ptN'),
        ('ptN','ptNext'),
        ('pta','ptag'),
        ('ptf','ptfirst'),
        ('ptj','ptjump'),
        ('ptl','ptlast'),
        ('ptn','ptn'),
        ('ptn','ptnext'),
        ('ptp','ptprevious'),
        ('ptr','ptrewind'),
        ('pts','ptselect'),
        ('pu','put'),
        ('pw','pwd'),
        ('py','py'),
        ('py','python'),
        ('py3','py3'),
        ('py3','py3'),
        ('py3do','py3do'),
        ('pydo','pydo'),
        ('pyf','pyfile'),
        ('python3','python3'),
        ('q','q'),
        ('q','quit'),
        ('qa','qall'),
        ('quita','quitall'),
        ('r','r'),
        ('r','read'),
        ('re','re'),
        ('rec','recover'),
        ('red','red'),
        ('red','redo'),
        ('redi','redir'),
        ('redr','redraw'),
        ('redraws','redrawstatus'),
        ('reg','registers'),
        ('res','resize'),
        ('ret','retab'),
        ('retu','return'),
        ('rew','rewind'),
        ('ri','right'),
        ('rightb','rightbelow'),
        ('ru','ru'),
        ('ru','runtime'),
        ('rub','ruby'),
        ('rubyd','rubydo'),
        ('rubyf','rubyfile'),
        ('rundo','rundo'),
        ('rv','rviminfo'),
        ('sN','sNext'),
        ('sa','sargument'),
        ('sal','sall'),
        ('san','sandbox'),
        ('sav','saveas'),
        ('sb','sbuffer'),
        ('sbN','sbNext'),
        ('sba','sball'),
        ('sbf','sbfirst'),
        ('sbl','sblast'),
        ('sbm','sbmodified'),
        ('sbn','sbnext'),
        ('sbp','sbprevious'),
        ('sbr','sbrewind'),
        ('scrip','scrip'),
        ('scrip','scriptnames'),
        ('scripte','scriptencoding'),
        ('scs','scs'),
        ('scscope','scscope'),
        ('se','set'),
        ('setf','setfiletype'),
        ('setg','setglobal'),
        ('setl','setlocal'),
        ('sf','sfind'),
        ('sfir','sfirst'),
        ('sh','shell'),
        ('si','si'),
        ('sig','sig'),
        ('sign','sign'),
        ('sil','silent'),
        ('sim','simalt'),
        ('sl','sl'),
        ('sl','sleep'),
        ('sla','slast'),
        ('sm','smagic'),
        ('sm','smap'),
        ('sme','sme'),
        ('smenu','smenu'),
        ('sn','snext'),
        ('sni','sniff'),
        ('sno','snomagic'),
        ('snoreme','snoreme'),
        ('snoremenu','snoremenu'),
        ('so','so'),
        ('so','source'),
        ('sor','sort'),
        ('sp','split'),
        ('spe','spe'),
        ('spe','spellgood'),
        ('spelld','spelldump'),
        ('spelli','spellinfo'),
        ('spellr','spellrepall'),
        ('spellu','spellundo'),
        ('spellw','spellwrong'),
        ('spr','sprevious'),
        ('sre','srewind'),
        ('st','st'),
        ('st','stop'),
        ('sta','stag'),
        ('star','star'),
        ('star','startinsert'),
        ('start','start'),
        ('startg','startgreplace'),
        ('startr','startreplace'),
        ('stj','stjump'),
        ('stopi','stopinsert'),
        ('sts','stselect'),
        ('sun','sunhide'),
        ('sunme','sunme'),
        ('sunmenu','sunmenu'),
        ('sus','suspend'),
        ('sv','sview'),
        ('sw','swapname'),
        ('sy','sy'),
        ('syn','syn'),
        ('sync','sync'),
        ('syncbind','syncbind'),
        ('syntime','syntime'),
        ('t','t'),
        ('tN','tN'),
        ('tN','tNext'),
        ('ta','ta'),
        ('ta','tag'),
        ('tab','tab'),
        ('tabN','tabN'),
        ('tabN','tabNext'),
        ('tabc','tabclose'),
        ('tabd','tabdo'),
        ('tabe','tabedit'),
        ('tabf','tabfind'),
        ('tabfir','tabfirst'),
        ('tabl','tablast'),
        ('tabm','tabmove'),
        ('tabn','tabnext'),
        ('tabnew','tabnew'),
        ('tabo','tabonly'),
        ('tabp','tabprevious'),
        ('tabr','tabrewind'),
        ('tabs','tabs'),
        ('tags','tags'),
        ('tc','tcl'),
        ('tcld','tcldo'),
        ('tclf','tclfile'),
        ('te','tearoff'),
        ('tf','tfirst'),
        ('th','throw'),
        ('tj','tjump'),
        ('tl','tlast'),
        ('tm','tm'),
        ('tm','tmenu'),
        ('tn','tn'),
        ('tn','tnext'),
        ('to','topleft'),
        ('tp','tprevious'),
        ('tr','tr'),
        ('tr','trewind'),
        ('try','try'),
        ('ts','tselect'),
        ('tu','tu'),
        ('tu','tunmenu'),
        ('u','u'),
        ('u','undo'),
        ('un','un'),
        ('una','unabbreviate'),
        ('undoj','undojoin'),
        ('undol','undolist'),
        ('unh','unhide'),
        ('unl','unl'),
        ('unlo','unlockvar'),
        ('uns','unsilent'),
        ('up','update'),
        ('v','v'),
        ('ve','ve'),
        ('ve','version'),
        ('verb','verbose'),
        ('vert','vertical'),
        ('vi','vi'),
        ('vi','visual'),
        ('vie','view'),
        ('vim','vimgrep'),
        ('vimgrepa','vimgrepadd'),
        ('viu','viusage'),
        ('vmapc','vmapclear'),
        ('vne','vnew'),
        ('vs','vsplit'),
        ('w','w'),
        ('w','write'),
        ('wN','wNext'),
        ('wa','wall'),
        ('wh','while'),
        ('win','win'),
        ('win','winsize'),
        ('winc','wincmd'),
        ('windo','windo'),
        ('winp','winpos'),
        ('wn','wnext'),
        ('wp','wprevious'),
        ('wq','wq'),
        ('wqa','wqall'),
        ('ws','wsverb'),
        ('wundo','wundo'),
        ('wv','wviminfo'),
        ('x','x'),
        ('x','xit'),
        ('xa','xall'),
        ('xmapc','xmapclear'),
        ('xme','xme'),
        ('xmenu','xmenu'),
        ('xnoreme','xnoreme'),
        ('xnoremenu','xnoremenu'),
        ('xunme','xunme'),
        ('xunmenu','xunmenu'),
        ('xwininfo','xwininfo'),
        ('y','yank'),
    )
    return var
command = _getcommand()

def _getoption():
    var = (
        ('acd','acd'),
        ('ai','ai'),
        ('akm','akm'),
        ('al','al'),
        ('aleph','aleph'),
        ('allowrevins','allowrevins'),
        ('altkeymap','altkeymap'),
        ('ambiwidth','ambiwidth'),
        ('ambw','ambw'),
        ('anti','anti'),
        ('antialias','antialias'),
        ('ar','ar'),
        ('arab','arab'),
        ('arabic','arabic'),
        ('arabicshape','arabicshape'),
        ('ari','ari'),
        ('arshape','arshape'),
        ('autochdir','autochdir'),
        ('autoindent','autoindent'),
        ('autoread','autoread'),
        ('autowrite','autowrite'),
        ('autowriteall','autowriteall'),
        ('aw','aw'),
        ('awa','awa'),
        ('background','background'),
        ('backspace','backspace'),
        ('backup','backup'),
        ('backupcopy','backupcopy'),
        ('backupdir','backupdir'),
        ('backupext','backupext'),
        ('backupskip','backupskip'),
        ('balloondelay','balloondelay'),
        ('ballooneval','ballooneval'),
        ('balloonexpr','balloonexpr'),
        ('bdir','bdir'),
        ('bdlay','bdlay'),
        ('beval','beval'),
        ('bex','bex'),
        ('bexpr','bexpr'),
        ('bg','bg'),
        ('bh','bh'),
        ('bin','bin'),
        ('binary','binary'),
        ('biosk','biosk'),
        ('bioskey','bioskey'),
        ('bk','bk'),
        ('bkc','bkc'),
        ('bl','bl'),
        ('bomb','bomb'),
        ('breakat','breakat'),
        ('brk','brk'),
        ('browsedir','browsedir'),
        ('bs','bs'),
        ('bsdir','bsdir'),
        ('bsk','bsk'),
        ('bt','bt'),
        ('bufhidden','bufhidden'),
        ('buflisted','buflisted'),
        ('buftype','buftype'),
        ('casemap','casemap'),
        ('cb','cb'),
        ('cc','cc'),
        ('ccv','ccv'),
        ('cd','cd'),
        ('cdpath','cdpath'),
        ('cedit','cedit'),
        ('cf','cf'),
        ('cfu','cfu'),
        ('ch','ch'),
        ('charconvert','charconvert'),
        ('ci','ci'),
        ('cin','cin'),
        ('cindent','cindent'),
        ('cink','cink'),
        ('cinkeys','cinkeys'),
        ('cino','cino'),
        ('cinoptions','cinoptions'),
        ('cinw','cinw'),
        ('cinwords','cinwords'),
        ('clipboard','clipboard'),
        ('cmdheight','cmdheight'),
        ('cmdwinheight','cmdwinheight'),
        ('cmp','cmp'),
        ('cms','cms'),
        ('co','co'),
        ('cocu','cocu'),
        ('cole','cole'),
        ('colorcolumn','colorcolumn'),
        ('columns','columns'),
        ('com','com'),
        ('comments','comments'),
        ('commentstring','commentstring'),
        ('compatible','compatible'),
        ('complete','complete'),
        ('completefunc','completefunc'),
        ('completeopt','completeopt'),
        ('concealcursor','concealcursor'),
        ('conceallevel','conceallevel'),
        ('confirm','confirm'),
        ('consk','consk'),
        ('conskey','conskey'),
        ('copyindent','copyindent'),
        ('cot','cot'),
        ('cp','cp'),
        ('cpo','cpo'),
        ('cpoptions','cpoptions'),
        ('cpt','cpt'),
        ('crb','crb'),
        ('cryptmethod','cryptmethod'),
        ('cscopepathcomp','cscopepathcomp'),
        ('cscopeprg','cscopeprg'),
        ('cscopequickfix','cscopequickfix'),
        ('cscoperelative','cscoperelative'),
        ('cscopetag','cscopetag'),
        ('cscopetagorder','cscopetagorder'),
        ('cscopeverbose','cscopeverbose'),
        ('cspc','cspc'),
        ('csprg','csprg'),
        ('csqf','csqf'),
        ('csre','csre'),
        ('cst','cst'),
        ('csto','csto'),
        ('csverb','csverb'),
        ('cuc','cuc'),
        ('cul','cul'),
        ('cursorbind','cursorbind'),
        ('cursorcolumn','cursorcolumn'),
        ('cursorline','cursorline'),
        ('cwh','cwh'),
        ('debug','debug'),
        ('deco','deco'),
        ('def','def'),
        ('define','define'),
        ('delcombine','delcombine'),
        ('dex','dex'),
        ('dg','dg'),
        ('dict','dict'),
        ('dictionary','dictionary'),
        ('diff','diff'),
        ('diffexpr','diffexpr'),
        ('diffopt','diffopt'),
        ('digraph','digraph'),
        ('dip','dip'),
        ('dir','dir'),
        ('directory','directory'),
        ('display','display'),
        ('dy','dy'),
        ('ea','ea'),
        ('ead','ead'),
        ('eadirection','eadirection'),
        ('eb','eb'),
        ('ed','ed'),
        ('edcompatible','edcompatible'),
        ('ef','ef'),
        ('efm','efm'),
        ('ei','ei'),
        ('ek','ek'),
        ('enc','enc'),
        ('encoding','encoding'),
        ('endofline','endofline'),
        ('eol','eol'),
        ('ep','ep'),
        ('equalalways','equalalways'),
        ('equalprg','equalprg'),
        ('errorbells','errorbells'),
        ('errorfile','errorfile'),
        ('errorformat','errorformat'),
        ('esckeys','esckeys'),
        ('et','et'),
        ('eventignore','eventignore'),
        ('ex','ex'),
        ('expandtab','expandtab'),
        ('exrc','exrc'),
        ('fcl','fcl'),
        ('fcs','fcs'),
        ('fdc','fdc'),
        ('fde','fde'),
        ('fdi','fdi'),
        ('fdl','fdl'),
        ('fdls','fdls'),
        ('fdm','fdm'),
        ('fdn','fdn'),
        ('fdo','fdo'),
        ('fdt','fdt'),
        ('fen','fen'),
        ('fenc','fenc'),
        ('fencs','fencs'),
        ('fex','fex'),
        ('ff','ff'),
        ('ffs','ffs'),
        ('fic','fic'),
        ('fileencoding','fileencoding'),
        ('fileencodings','fileencodings'),
        ('fileformat','fileformat'),
        ('fileformats','fileformats'),
        ('fileignorecase','fileignorecase'),
        ('filetype','filetype'),
        ('fillchars','fillchars'),
        ('fk','fk'),
        ('fkmap','fkmap'),
        ('flp','flp'),
        ('fml','fml'),
        ('fmr','fmr'),
        ('fo','fo'),
        ('foldclose','foldclose'),
        ('foldcolumn','foldcolumn'),
        ('foldenable','foldenable'),
        ('foldexpr','foldexpr'),
        ('foldignore','foldignore'),
        ('foldlevel','foldlevel'),
        ('foldlevelstart','foldlevelstart'),
        ('foldmarker','foldmarker'),
        ('foldmethod','foldmethod'),
        ('foldminlines','foldminlines'),
        ('foldnestmax','foldnestmax'),
        ('foldopen','foldopen'),
        ('foldtext','foldtext'),
        ('formatexpr','formatexpr'),
        ('formatlistpat','formatlistpat'),
        ('formatoptions','formatoptions'),
        ('formatprg','formatprg'),
        ('fp','fp'),
        ('fs','fs'),
        ('fsync','fsync'),
        ('ft','ft'),
        ('gcr','gcr'),
        ('gd','gd'),
        ('gdefault','gdefault'),
        ('gfm','gfm'),
        ('gfn','gfn'),
        ('gfs','gfs'),
        ('gfw','gfw'),
        ('ghr','ghr'),
        ('go','go'),
        ('gp','gp'),
        ('grepformat','grepformat'),
        ('grepprg','grepprg'),
        ('gtl','gtl'),
        ('gtt','gtt'),
        ('guicursor','guicursor'),
        ('guifont','guifont'),
        ('guifontset','guifontset'),
        ('guifontwide','guifontwide'),
        ('guiheadroom','guiheadroom'),
        ('guioptions','guioptions'),
        ('guipty','guipty'),
        ('guitablabel','guitablabel'),
        ('guitabtooltip','guitabtooltip'),
        ('helpfile','helpfile'),
        ('helpheight','helpheight'),
        ('helplang','helplang'),
        ('hf','hf'),
        ('hh','hh'),
        ('hi','hi'),
        ('hid','hid'),
        ('hidden','hidden'),
        ('highlight','highlight'),
        ('history','history'),
        ('hk','hk'),
        ('hkmap','hkmap'),
        ('hkmapp','hkmapp'),
        ('hkp','hkp'),
        ('hl','hl'),
        ('hlg','hlg'),
        ('hls','hls'),
        ('hlsearch','hlsearch'),
        ('ic','ic'),
        ('icon','icon'),
        ('iconstring','iconstring'),
        ('ignorecase','ignorecase'),
        ('im','im'),
        ('imactivatefunc','imactivatefunc'),
        ('imactivatekey','imactivatekey'),
        ('imaf','imaf'),
        ('imak','imak'),
        ('imc','imc'),
        ('imcmdline','imcmdline'),
        ('imd','imd'),
        ('imdisable','imdisable'),
        ('imi','imi'),
        ('iminsert','iminsert'),
        ('ims','ims'),
        ('imsearch','imsearch'),
        ('imsf','imsf'),
        ('imstatusfunc','imstatusfunc'),
        ('inc','inc'),
        ('include','include'),
        ('includeexpr','includeexpr'),
        ('incsearch','incsearch'),
        ('inde','inde'),
        ('indentexpr','indentexpr'),
        ('indentkeys','indentkeys'),
        ('indk','indk'),
        ('inex','inex'),
        ('inf','inf'),
        ('infercase','infercase'),
        ('inoremap','inoremap'),
        ('insertmode','insertmode'),
        ('invacd','invacd'),
        ('invai','invai'),
        ('invakm','invakm'),
        ('invallowrevins','invallowrevins'),
        ('invaltkeymap','invaltkeymap'),
        ('invanti','invanti'),
        ('invantialias','invantialias'),
        ('invar','invar'),
        ('invarab','invarab'),
        ('invarabic','invarabic'),
        ('invarabicshape','invarabicshape'),
        ('invari','invari'),
        ('invarshape','invarshape'),
        ('invautochdir','invautochdir'),
        ('invautoindent','invautoindent'),
        ('invautoread','invautoread'),
        ('invautowrite','invautowrite'),
        ('invautowriteall','invautowriteall'),
        ('invaw','invaw'),
        ('invawa','invawa'),
        ('invbackup','invbackup'),
        ('invballooneval','invballooneval'),
        ('invbeval','invbeval'),
        ('invbin','invbin'),
        ('invbinary','invbinary'),
        ('invbiosk','invbiosk'),
        ('invbioskey','invbioskey'),
        ('invbk','invbk'),
        ('invbl','invbl'),
        ('invbomb','invbomb'),
        ('invbuflisted','invbuflisted'),
        ('invcf','invcf'),
        ('invci','invci'),
        ('invcin','invcin'),
        ('invcindent','invcindent'),
        ('invcompatible','invcompatible'),
        ('invconfirm','invconfirm'),
        ('invconsk','invconsk'),
        ('invconskey','invconskey'),
        ('invcopyindent','invcopyindent'),
        ('invcp','invcp'),
        ('invcrb','invcrb'),
        ('invcscoperelative','invcscoperelative'),
        ('invcscopetag','invcscopetag'),
        ('invcscopeverbose','invcscopeverbose'),
        ('invcsre','invcsre'),
        ('invcst','invcst'),
        ('invcsverb','invcsverb'),
        ('invcuc','invcuc'),
        ('invcul','invcul'),
        ('invcursorbind','invcursorbind'),
        ('invcursorcolumn','invcursorcolumn'),
        ('invcursorline','invcursorline'),
        ('invdeco','invdeco'),
        ('invdelcombine','invdelcombine'),
        ('invdg','invdg'),
        ('invdiff','invdiff'),
        ('invdigraph','invdigraph'),
        ('invea','invea'),
        ('inveb','inveb'),
        ('inved','inved'),
        ('invedcompatible','invedcompatible'),
        ('invek','invek'),
        ('invendofline','invendofline'),
        ('inveol','inveol'),
        ('invequalalways','invequalalways'),
        ('inverrorbells','inverrorbells'),
        ('invesckeys','invesckeys'),
        ('invet','invet'),
        ('invex','invex'),
        ('invexpandtab','invexpandtab'),
        ('invexrc','invexrc'),
        ('invfen','invfen'),
        ('invfic','invfic'),
        ('invfileignorecase','invfileignorecase'),
        ('invfk','invfk'),
        ('invfkmap','invfkmap'),
        ('invfoldenable','invfoldenable'),
        ('invgd','invgd'),
        ('invgdefault','invgdefault'),
        ('invguipty','invguipty'),
        ('invhid','invhid'),
        ('invhidden','invhidden'),
        ('invhk','invhk'),
        ('invhkmap','invhkmap'),
        ('invhkmapp','invhkmapp'),
        ('invhkp','invhkp'),
        ('invhls','invhls'),
        ('invhlsearch','invhlsearch'),
        ('invic','invic'),
        ('invicon','invicon'),
        ('invignorecase','invignorecase'),
        ('invim','invim'),
        ('invimc','invimc'),
        ('invimcmdline','invimcmdline'),
        ('invimd','invimd'),
        ('invimdisable','invimdisable'),
        ('invincsearch','invincsearch'),
        ('invinf','invinf'),
        ('invinfercase','invinfercase'),
        ('invinsertmode','invinsertmode'),
        ('invis','invis'),
        ('invjoinspaces','invjoinspaces'),
        ('invjs','invjs'),
        ('invlazyredraw','invlazyredraw'),
        ('invlbr','invlbr'),
        ('invlinebreak','invlinebreak'),
        ('invlisp','invlisp'),
        ('invlist','invlist'),
        ('invloadplugins','invloadplugins'),
        ('invlpl','invlpl'),
        ('invlz','invlz'),
        ('invma','invma'),
        ('invmacatsui','invmacatsui'),
        ('invmagic','invmagic'),
        ('invmh','invmh'),
        ('invml','invml'),
        ('invmod','invmod'),
        ('invmodeline','invmodeline'),
        ('invmodifiable','invmodifiable'),
        ('invmodified','invmodified'),
        ('invmore','invmore'),
        ('invmousef','invmousef'),
        ('invmousefocus','invmousefocus'),
        ('invmousehide','invmousehide'),
        ('invnu','invnu'),
        ('invnumber','invnumber'),
        ('invodev','invodev'),
        ('invopendevice','invopendevice'),
        ('invpaste','invpaste'),
        ('invpi','invpi'),
        ('invpreserveindent','invpreserveindent'),
        ('invpreviewwindow','invpreviewwindow'),
        ('invprompt','invprompt'),
        ('invpvw','invpvw'),
        ('invreadonly','invreadonly'),
        ('invrelativenumber','invrelativenumber'),
        ('invremap','invremap'),
        ('invrestorescreen','invrestorescreen'),
        ('invrevins','invrevins'),
        ('invri','invri'),
        ('invrightleft','invrightleft'),
        ('invrl','invrl'),
        ('invrnu','invrnu'),
        ('invro','invro'),
        ('invrs','invrs'),
        ('invru','invru'),
        ('invruler','invruler'),
        ('invsb','invsb'),
        ('invsc','invsc'),
        ('invscb','invscb'),
        ('invscrollbind','invscrollbind'),
        ('invscs','invscs'),
        ('invsecure','invsecure'),
        ('invsft','invsft'),
        ('invshellslash','invshellslash'),
        ('invshelltemp','invshelltemp'),
        ('invshiftround','invshiftround'),
        ('invshortname','invshortname'),
        ('invshowcmd','invshowcmd'),
        ('invshowfulltag','invshowfulltag'),
        ('invshowmatch','invshowmatch'),
        ('invshowmode','invshowmode'),
        ('invsi','invsi'),
        ('invsm','invsm'),
        ('invsmartcase','invsmartcase'),
        ('invsmartindent','invsmartindent'),
        ('invsmarttab','invsmarttab'),
        ('invsmd','invsmd'),
        ('invsn','invsn'),
        ('invsol','invsol'),
        ('invspell','invspell'),
        ('invsplitbelow','invsplitbelow'),
        ('invsplitright','invsplitright'),
        ('invspr','invspr'),
        ('invsr','invsr'),
        ('invssl','invssl'),
        ('invsta','invsta'),
        ('invstartofline','invstartofline'),
        ('invstmp','invstmp'),
        ('invswapfile','invswapfile'),
        ('invswf','invswf'),
        ('invta','invta'),
        ('invtagbsearch','invtagbsearch'),
        ('invtagrelative','invtagrelative'),
        ('invtagstack','invtagstack'),
        ('invtbi','invtbi'),
        ('invtbidi','invtbidi'),
        ('invtbs','invtbs'),
        ('invtermbidi','invtermbidi'),
        ('invterse','invterse'),
        ('invtextauto','invtextauto'),
        ('invtextmode','invtextmode'),
        ('invtf','invtf'),
        ('invtgst','invtgst'),
        ('invtildeop','invtildeop'),
        ('invtimeout','invtimeout'),
        ('invtitle','invtitle'),
        ('invto','invto'),
        ('invtop','invtop'),
        ('invtr','invtr'),
        ('invttimeout','invttimeout'),
        ('invttybuiltin','invttybuiltin'),
        ('invttyfast','invttyfast'),
        ('invtx','invtx'),
        ('invudf','invudf'),
        ('invundofile','invundofile'),
        ('invvb','invvb'),
        ('invvisualbell','invvisualbell'),
        ('invwa','invwa'),
        ('invwarn','invwarn'),
        ('invwb','invwb'),
        ('invweirdinvert','invweirdinvert'),
        ('invwfh','invwfh'),
        ('invwfw','invwfw'),
        ('invwic','invwic'),
        ('invwildignorecase','invwildignorecase'),
        ('invwildmenu','invwildmenu'),
        ('invwinfixheight','invwinfixheight'),
        ('invwinfixwidth','invwinfixwidth'),
        ('invwiv','invwiv'),
        ('invwmnu','invwmnu'),
        ('invwrap','invwrap'),
        ('invwrapscan','invwrapscan'),
        ('invwrite','invwrite'),
        ('invwriteany','invwriteany'),
        ('invwritebackup','invwritebackup'),
        ('invws','invws'),
        ('is','is'),
        ('isf','isf'),
        ('isfname','isfname'),
        ('isi','isi'),
        ('isident','isident'),
        ('isk','isk'),
        ('iskeyword','iskeyword'),
        ('isp','isp'),
        ('isprint','isprint'),
        ('joinspaces','joinspaces'),
        ('js','js'),
        ('key','key'),
        ('keymap','keymap'),
        ('keymodel','keymodel'),
        ('keywordprg','keywordprg'),
        ('km','km'),
        ('kmp','kmp'),
        ('kp','kp'),
        ('langmap','langmap'),
        ('langmenu','langmenu'),
        ('laststatus','laststatus'),
        ('lazyredraw','lazyredraw'),
        ('lbr','lbr'),
        ('lcs','lcs'),
        ('linebreak','linebreak'),
        ('lines','lines'),
        ('linespace','linespace'),
        ('lisp','lisp'),
        ('lispwords','lispwords'),
        ('list','list'),
        ('listchars','listchars'),
        ('lm','lm'),
        ('lmap','lmap'),
        ('loadplugins','loadplugins'),
        ('lpl','lpl'),
        ('ls','ls'),
        ('lsp','lsp'),
        ('lw','lw'),
        ('lz','lz'),
        ('ma','ma'),
        ('macatsui','macatsui'),
        ('magic','magic'),
        ('makeef','makeef'),
        ('makeprg','makeprg'),
        ('mat','mat'),
        ('matchpairs','matchpairs'),
        ('matchtime','matchtime'),
        ('maxcombine','maxcombine'),
        ('maxfuncdepth','maxfuncdepth'),
        ('maxmapdepth','maxmapdepth'),
        ('maxmem','maxmem'),
        ('maxmempattern','maxmempattern'),
        ('maxmemtot','maxmemtot'),
        ('mco','mco'),
        ('mef','mef'),
        ('menuitems','menuitems'),
        ('mfd','mfd'),
        ('mh','mh'),
        ('mis','mis'),
        ('mkspellmem','mkspellmem'),
        ('ml','ml'),
        ('mls','mls'),
        ('mm','mm'),
        ('mmd','mmd'),
        ('mmp','mmp'),
        ('mmt','mmt'),
        ('mod','mod'),
        ('modeline','modeline'),
        ('modelines','modelines'),
        ('modifiable','modifiable'),
        ('modified','modified'),
        ('more','more'),
        ('mouse','mouse'),
        ('mousef','mousef'),
        ('mousefocus','mousefocus'),
        ('mousehide','mousehide'),
        ('mousem','mousem'),
        ('mousemodel','mousemodel'),
        ('mouses','mouses'),
        ('mouseshape','mouseshape'),
        ('mouset','mouset'),
        ('mousetime','mousetime'),
        ('mp','mp'),
        ('mps','mps'),
        ('msm','msm'),
        ('mzq','mzq'),
        ('mzquantum','mzquantum'),
        ('nf','nf'),
        ('nnoremap','nnoremap'),
        ('noacd','noacd'),
        ('noai','noai'),
        ('noakm','noakm'),
        ('noallowrevins','noallowrevins'),
        ('noaltkeymap','noaltkeymap'),
        ('noanti','noanti'),
        ('noantialias','noantialias'),
        ('noar','noar'),
        ('noarab','noarab'),
        ('noarabic','noarabic'),
        ('noarabicshape','noarabicshape'),
        ('noari','noari'),
        ('noarshape','noarshape'),
        ('noautochdir','noautochdir'),
        ('noautoindent','noautoindent'),
        ('noautoread','noautoread'),
        ('noautowrite','noautowrite'),
        ('noautowriteall','noautowriteall'),
        ('noaw','noaw'),
        ('noawa','noawa'),
        ('nobackup','nobackup'),
        ('noballooneval','noballooneval'),
        ('nobeval','nobeval'),
        ('nobin','nobin'),
        ('nobinary','nobinary'),
        ('nobiosk','nobiosk'),
        ('nobioskey','nobioskey'),
        ('nobk','nobk'),
        ('nobl','nobl'),
        ('nobomb','nobomb'),
        ('nobuflisted','nobuflisted'),
        ('nocf','nocf'),
        ('noci','noci'),
        ('nocin','nocin'),
        ('nocindent','nocindent'),
        ('nocompatible','nocompatible'),
        ('noconfirm','noconfirm'),
        ('noconsk','noconsk'),
        ('noconskey','noconskey'),
        ('nocopyindent','nocopyindent'),
        ('nocp','nocp'),
        ('nocrb','nocrb'),
        ('nocscoperelative','nocscoperelative'),
        ('nocscopetag','nocscopetag'),
        ('nocscopeverbose','nocscopeverbose'),
        ('nocsre','nocsre'),
        ('nocst','nocst'),
        ('nocsverb','nocsverb'),
        ('nocuc','nocuc'),
        ('nocul','nocul'),
        ('nocursorbind','nocursorbind'),
        ('nocursorcolumn','nocursorcolumn'),
        ('nocursorline','nocursorline'),
        ('nodeco','nodeco'),
        ('nodelcombine','nodelcombine'),
        ('nodg','nodg'),
        ('nodiff','nodiff'),
        ('nodigraph','nodigraph'),
        ('noea','noea'),
        ('noeb','noeb'),
        ('noed','noed'),
        ('noedcompatible','noedcompatible'),
        ('noek','noek'),
        ('noendofline','noendofline'),
        ('noeol','noeol'),
        ('noequalalways','noequalalways'),
        ('noerrorbells','noerrorbells'),
        ('noesckeys','noesckeys'),
        ('noet','noet'),
        ('noex','noex'),
        ('noexpandtab','noexpandtab'),
        ('noexrc','noexrc'),
        ('nofen','nofen'),
        ('nofic','nofic'),
        ('nofileignorecase','nofileignorecase'),
        ('nofk','nofk'),
        ('nofkmap','nofkmap'),
        ('nofoldenable','nofoldenable'),
        ('nogd','nogd'),
        ('nogdefault','nogdefault'),
        ('noguipty','noguipty'),
        ('nohid','nohid'),
        ('nohidden','nohidden'),
        ('nohk','nohk'),
        ('nohkmap','nohkmap'),
        ('nohkmapp','nohkmapp'),
        ('nohkp','nohkp'),
        ('nohls','nohls'),
        ('nohlsearch','nohlsearch'),
        ('noic','noic'),
        ('noicon','noicon'),
        ('noignorecase','noignorecase'),
        ('noim','noim'),
        ('noimc','noimc'),
        ('noimcmdline','noimcmdline'),
        ('noimd','noimd'),
        ('noimdisable','noimdisable'),
        ('noincsearch','noincsearch'),
        ('noinf','noinf'),
        ('noinfercase','noinfercase'),
        ('noinsertmode','noinsertmode'),
        ('nois','nois'),
        ('nojoinspaces','nojoinspaces'),
        ('nojs','nojs'),
        ('nolazyredraw','nolazyredraw'),
        ('nolbr','nolbr'),
        ('nolinebreak','nolinebreak'),
        ('nolisp','nolisp'),
        ('nolist','nolist'),
        ('noloadplugins','noloadplugins'),
        ('nolpl','nolpl'),
        ('nolz','nolz'),
        ('noma','noma'),
        ('nomacatsui','nomacatsui'),
        ('nomagic','nomagic'),
        ('nomh','nomh'),
        ('noml','noml'),
        ('nomod','nomod'),
        ('nomodeline','nomodeline'),
        ('nomodifiable','nomodifiable'),
        ('nomodified','nomodified'),
        ('nomore','nomore'),
        ('nomousef','nomousef'),
        ('nomousefocus','nomousefocus'),
        ('nomousehide','nomousehide'),
        ('nonu','nonu'),
        ('nonumber','nonumber'),
        ('noodev','noodev'),
        ('noopendevice','noopendevice'),
        ('nopaste','nopaste'),
        ('nopi','nopi'),
        ('nopreserveindent','nopreserveindent'),
        ('nopreviewwindow','nopreviewwindow'),
        ('noprompt','noprompt'),
        ('nopvw','nopvw'),
        ('noreadonly','noreadonly'),
        ('norelativenumber','norelativenumber'),
        ('noremap','noremap'),
        ('norestorescreen','norestorescreen'),
        ('norevins','norevins'),
        ('nori','nori'),
        ('norightleft','norightleft'),
        ('norl','norl'),
        ('nornu','nornu'),
        ('noro','noro'),
        ('nors','nors'),
        ('noru','noru'),
        ('noruler','noruler'),
        ('nosb','nosb'),
        ('nosc','nosc'),
        ('noscb','noscb'),
        ('noscrollbind','noscrollbind'),
        ('noscs','noscs'),
        ('nosecure','nosecure'),
        ('nosft','nosft'),
        ('noshellslash','noshellslash'),
        ('noshelltemp','noshelltemp'),
        ('noshiftround','noshiftround'),
        ('noshortname','noshortname'),
        ('noshowcmd','noshowcmd'),
        ('noshowfulltag','noshowfulltag'),
        ('noshowmatch','noshowmatch'),
        ('noshowmode','noshowmode'),
        ('nosi','nosi'),
        ('nosm','nosm'),
        ('nosmartcase','nosmartcase'),
        ('nosmartindent','nosmartindent'),
        ('nosmarttab','nosmarttab'),
        ('nosmd','nosmd'),
        ('nosn','nosn'),
        ('nosol','nosol'),
        ('nospell','nospell'),
        ('nosplitbelow','nosplitbelow'),
        ('nosplitright','nosplitright'),
        ('nospr','nospr'),
        ('nosr','nosr'),
        ('nossl','nossl'),
        ('nosta','nosta'),
        ('nostartofline','nostartofline'),
        ('nostmp','nostmp'),
        ('noswapfile','noswapfile'),
        ('noswf','noswf'),
        ('nota','nota'),
        ('notagbsearch','notagbsearch'),
        ('notagrelative','notagrelative'),
        ('notagstack','notagstack'),
        ('notbi','notbi'),
        ('notbidi','notbidi'),
        ('notbs','notbs'),
        ('notermbidi','notermbidi'),
        ('noterse','noterse'),
        ('notextauto','notextauto'),
        ('notextmode','notextmode'),
        ('notf','notf'),
        ('notgst','notgst'),
        ('notildeop','notildeop'),
        ('notimeout','notimeout'),
        ('notitle','notitle'),
        ('noto','noto'),
        ('notop','notop'),
        ('notr','notr'),
        ('nottimeout','nottimeout'),
        ('nottybuiltin','nottybuiltin'),
        ('nottyfast','nottyfast'),
        ('notx','notx'),
        ('noudf','noudf'),
        ('noundofile','noundofile'),
        ('novb','novb'),
        ('novisualbell','novisualbell'),
        ('nowa','nowa'),
        ('nowarn','nowarn'),
        ('nowb','nowb'),
        ('noweirdinvert','noweirdinvert'),
        ('nowfh','nowfh'),
        ('nowfw','nowfw'),
        ('nowic','nowic'),
        ('nowildignorecase','nowildignorecase'),
        ('nowildmenu','nowildmenu'),
        ('nowinfixheight','nowinfixheight'),
        ('nowinfixwidth','nowinfixwidth'),
        ('nowiv','nowiv'),
        ('nowmnu','nowmnu'),
        ('nowrap','nowrap'),
        ('nowrapscan','nowrapscan'),
        ('nowrite','nowrite'),
        ('nowriteany','nowriteany'),
        ('nowritebackup','nowritebackup'),
        ('nows','nows'),
        ('nrformats','nrformats'),
        ('nu','nu'),
        ('number','number'),
        ('numberwidth','numberwidth'),
        ('nuw','nuw'),
        ('odev','odev'),
        ('oft','oft'),
        ('ofu','ofu'),
        ('omnifunc','omnifunc'),
        ('opendevice','opendevice'),
        ('operatorfunc','operatorfunc'),
        ('opfunc','opfunc'),
        ('osfiletype','osfiletype'),
        ('pa','pa'),
        ('para','para'),
        ('paragraphs','paragraphs'),
        ('paste','paste'),
        ('pastetoggle','pastetoggle'),
        ('patchexpr','patchexpr'),
        ('patchmode','patchmode'),
        ('path','path'),
        ('pdev','pdev'),
        ('penc','penc'),
        ('pex','pex'),
        ('pexpr','pexpr'),
        ('pfn','pfn'),
        ('ph','ph'),
        ('pheader','pheader'),
        ('pi','pi'),
        ('pm','pm'),
        ('pmbcs','pmbcs'),
        ('pmbfn','pmbfn'),
        ('popt','popt'),
        ('preserveindent','preserveindent'),
        ('previewheight','previewheight'),
        ('previewwindow','previewwindow'),
        ('printdevice','printdevice'),
        ('printencoding','printencoding'),
        ('printexpr','printexpr'),
        ('printfont','printfont'),
        ('printheader','printheader'),
        ('printmbcharset','printmbcharset'),
        ('printmbfont','printmbfont'),
        ('printoptions','printoptions'),
        ('prompt','prompt'),
        ('pt','pt'),
        ('pumheight','pumheight'),
        ('pvh','pvh'),
        ('pvw','pvw'),
        ('qe','qe'),
        ('quoteescape','quoteescape'),
        ('rdt','rdt'),
        ('re','re'),
        ('readonly','readonly'),
        ('redrawtime','redrawtime'),
        ('regexpengine','regexpengine'),
        ('relativenumber','relativenumber'),
        ('remap','remap'),
        ('report','report'),
        ('restorescreen','restorescreen'),
        ('revins','revins'),
        ('ri','ri'),
        ('rightleft','rightleft'),
        ('rightleftcmd','rightleftcmd'),
        ('rl','rl'),
        ('rlc','rlc'),
        ('rnu','rnu'),
        ('ro','ro'),
        ('rs','rs'),
        ('rtp','rtp'),
        ('ru','ru'),
        ('ruf','ruf'),
        ('ruler','ruler'),
        ('rulerformat','rulerformat'),
        ('runtimepath','runtimepath'),
        ('sb','sb'),
        ('sbo','sbo'),
        ('sbr','sbr'),
        ('sc','sc'),
        ('scb','scb'),
        ('scr','scr'),
        ('scroll','scroll'),
        ('scrollbind','scrollbind'),
        ('scrolljump','scrolljump'),
        ('scrolloff','scrolloff'),
        ('scrollopt','scrollopt'),
        ('scs','scs'),
        ('sect','sect'),
        ('sections','sections'),
        ('secure','secure'),
        ('sel','sel'),
        ('selection','selection'),
        ('selectmode','selectmode'),
        ('sessionoptions','sessionoptions'),
        ('sft','sft'),
        ('sh','sh'),
        ('shcf','shcf'),
        ('shell','shell'),
        ('shellcmdflag','shellcmdflag'),
        ('shellpipe','shellpipe'),
        ('shellquote','shellquote'),
        ('shellredir','shellredir'),
        ('shellslash','shellslash'),
        ('shelltemp','shelltemp'),
        ('shelltype','shelltype'),
        ('shellxescape','shellxescape'),
        ('shellxquote','shellxquote'),
        ('shiftround','shiftround'),
        ('shiftwidth','shiftwidth'),
        ('shm','shm'),
        ('shortmess','shortmess'),
        ('shortname','shortname'),
        ('showbreak','showbreak'),
        ('showcmd','showcmd'),
        ('showfulltag','showfulltag'),
        ('showmatch','showmatch'),
        ('showmode','showmode'),
        ('showtabline','showtabline'),
        ('shq','shq'),
        ('si','si'),
        ('sidescroll','sidescroll'),
        ('sidescrolloff','sidescrolloff'),
        ('siso','siso'),
        ('sj','sj'),
        ('slm','slm'),
        ('sm','sm'),
        ('smartcase','smartcase'),
        ('smartindent','smartindent'),
        ('smarttab','smarttab'),
        ('smc','smc'),
        ('smd','smd'),
        ('sn','sn'),
        ('so','so'),
        ('softtabstop','softtabstop'),
        ('sol','sol'),
        ('sp','sp'),
        ('spc','spc'),
        ('spell','spell'),
        ('spellcapcheck','spellcapcheck'),
        ('spellfile','spellfile'),
        ('spelllang','spelllang'),
        ('spellsuggest','spellsuggest'),
        ('spf','spf'),
        ('spl','spl'),
        ('splitbelow','splitbelow'),
        ('splitright','splitright'),
        ('spr','spr'),
        ('sps','sps'),
        ('sr','sr'),
        ('srr','srr'),
        ('ss','ss'),
        ('ssl','ssl'),
        ('ssop','ssop'),
        ('st','st'),
        ('sta','sta'),
        ('stal','stal'),
        ('startofline','startofline'),
        ('statusline','statusline'),
        ('stl','stl'),
        ('stmp','stmp'),
        ('sts','sts'),
        ('su','su'),
        ('sua','sua'),
        ('suffixes','suffixes'),
        ('suffixesadd','suffixesadd'),
        ('sw','sw'),
        ('swapfile','swapfile'),
        ('swapsync','swapsync'),
        ('swb','swb'),
        ('swf','swf'),
        ('switchbuf','switchbuf'),
        ('sws','sws'),
        ('sxe','sxe'),
        ('sxq','sxq'),
        ('syn','syn'),
        ('synmaxcol','synmaxcol'),
        ('syntax','syntax'),
        ('t_AB','t_AB'),
        ('t_AF','t_AF'),
        ('t_AL','t_AL'),
        ('t_CS','t_CS'),
        ('t_CV','t_CV'),
        ('t_Ce','t_Ce'),
        ('t_Co','t_Co'),
        ('t_Cs','t_Cs'),
        ('t_DL','t_DL'),
        ('t_EI','t_EI'),
        ('t_F1','t_F1'),
        ('t_F2','t_F2'),
        ('t_F3','t_F3'),
        ('t_F4','t_F4'),
        ('t_F5','t_F5'),
        ('t_F6','t_F6'),
        ('t_F7','t_F7'),
        ('t_F8','t_F8'),
        ('t_F9','t_F9'),
        ('t_IE','t_IE'),
        ('t_IS','t_IS'),
        ('t_K1','t_K1'),
        ('t_K3','t_K3'),
        ('t_K4','t_K4'),
        ('t_K5','t_K5'),
        ('t_K6','t_K6'),
        ('t_K7','t_K7'),
        ('t_K8','t_K8'),
        ('t_K9','t_K9'),
        ('t_KA','t_KA'),
        ('t_KB','t_KB'),
        ('t_KC','t_KC'),
        ('t_KD','t_KD'),
        ('t_KE','t_KE'),
        ('t_KF','t_KF'),
        ('t_KG','t_KG'),
        ('t_KH','t_KH'),
        ('t_KI','t_KI'),
        ('t_KJ','t_KJ'),
        ('t_KK','t_KK'),
        ('t_KL','t_KL'),
        ('t_RI','t_RI'),
        ('t_RV','t_RV'),
        ('t_SI','t_SI'),
        ('t_Sb','t_Sb'),
        ('t_Sf','t_Sf'),
        ('t_WP','t_WP'),
        ('t_WS','t_WS'),
        ('t_ZH','t_ZH'),
        ('t_ZR','t_ZR'),
        ('t_al','t_al'),
        ('t_bc','t_bc'),
        ('t_cd','t_cd'),
        ('t_ce','t_ce'),
        ('t_cl','t_cl'),
        ('t_cm','t_cm'),
        ('t_cs','t_cs'),
        ('t_da','t_da'),
        ('t_db','t_db'),
        ('t_dl','t_dl'),
        ('t_fs','t_fs'),
        ('t_k1','t_k1'),
        ('t_k2','t_k2'),
        ('t_k3','t_k3'),
        ('t_k4','t_k4'),
        ('t_k5','t_k5'),
        ('t_k6','t_k6'),
        ('t_k7','t_k7'),
        ('t_k8','t_k8'),
        ('t_k9','t_k9'),
        ('t_kB','t_kB'),
        ('t_kD','t_kD'),
        ('t_kI','t_kI'),
        ('t_kN','t_kN'),
        ('t_kP','t_kP'),
        ('t_kb','t_kb'),
        ('t_kd','t_kd'),
        ('t_ke','t_ke'),
        ('t_kh','t_kh'),
        ('t_kl','t_kl'),
        ('t_kr','t_kr'),
        ('t_ks','t_ks'),
        ('t_ku','t_ku'),
        ('t_le','t_le'),
        ('t_mb','t_mb'),
        ('t_md','t_md'),
        ('t_me','t_me'),
        ('t_mr','t_mr'),
        ('t_ms','t_ms'),
        ('t_nd','t_nd'),
        ('t_op','t_op'),
        ('t_se','t_se'),
        ('t_so','t_so'),
        ('t_sr','t_sr'),
        ('t_te','t_te'),
        ('t_ti','t_ti'),
        ('t_ts','t_ts'),
        ('t_u7','t_u7'),
        ('t_ue','t_ue'),
        ('t_us','t_us'),
        ('t_ut','t_ut'),
        ('t_vb','t_vb'),
        ('t_ve','t_ve'),
        ('t_vi','t_vi'),
        ('t_vs','t_vs'),
        ('t_xs','t_xs'),
        ('ta','ta'),
        ('tabline','tabline'),
        ('tabpagemax','tabpagemax'),
        ('tabstop','tabstop'),
        ('tag','tag'),
        ('tagbsearch','tagbsearch'),
        ('taglength','taglength'),
        ('tagrelative','tagrelative'),
        ('tags','tags'),
        ('tagstack','tagstack'),
        ('tal','tal'),
        ('tb','tb'),
        ('tbi','tbi'),
        ('tbidi','tbidi'),
        ('tbis','tbis'),
        ('tbs','tbs'),
        ('tenc','tenc'),
        ('term','term'),
        ('termbidi','termbidi'),
        ('termencoding','termencoding'),
        ('terse','terse'),
        ('textauto','textauto'),
        ('textmode','textmode'),
        ('textwidth','textwidth'),
        ('tf','tf'),
        ('tgst','tgst'),
        ('thesaurus','thesaurus'),
        ('tildeop','tildeop'),
        ('timeout','timeout'),
        ('timeoutlen','timeoutlen'),
        ('title','title'),
        ('titlelen','titlelen'),
        ('titleold','titleold'),
        ('titlestring','titlestring'),
        ('tl','tl'),
        ('tm','tm'),
        ('to','to'),
        ('toolbar','toolbar'),
        ('toolbariconsize','toolbariconsize'),
        ('top','top'),
        ('tpm','tpm'),
        ('tr','tr'),
        ('ts','ts'),
        ('tsl','tsl'),
        ('tsr','tsr'),
        ('ttimeout','ttimeout'),
        ('ttimeoutlen','ttimeoutlen'),
        ('ttm','ttm'),
        ('tty','tty'),
        ('ttybuiltin','ttybuiltin'),
        ('ttyfast','ttyfast'),
        ('ttym','ttym'),
        ('ttymouse','ttymouse'),
        ('ttyscroll','ttyscroll'),
        ('ttytype','ttytype'),
        ('tw','tw'),
        ('tx','tx'),
        ('uc','uc'),
        ('udf','udf'),
        ('udir','udir'),
        ('ul','ul'),
        ('undodir','undodir'),
        ('undofile','undofile'),
        ('undolevels','undolevels'),
        ('undoreload','undoreload'),
        ('updatecount','updatecount'),
        ('updatetime','updatetime'),
        ('ur','ur'),
        ('ut','ut'),
        ('vb','vb'),
        ('vbs','vbs'),
        ('vdir','vdir'),
        ('ve','ve'),
        ('verbose','verbose'),
        ('verbosefile','verbosefile'),
        ('vfile','vfile'),
        ('vi','vi'),
        ('viewdir','viewdir'),
        ('viewoptions','viewoptions'),
        ('viminfo','viminfo'),
        ('virtualedit','virtualedit'),
        ('visualbell','visualbell'),
        ('vnoremap','vnoremap'),
        ('vop','vop'),
        ('wa','wa'),
        ('wak','wak'),
        ('warn','warn'),
        ('wb','wb'),
        ('wc','wc'),
        ('wcm','wcm'),
        ('wd','wd'),
        ('weirdinvert','weirdinvert'),
        ('wfh','wfh'),
        ('wfw','wfw'),
        ('wh','wh'),
        ('whichwrap','whichwrap'),
        ('wi','wi'),
        ('wic','wic'),
        ('wig','wig'),
        ('wildchar','wildchar'),
        ('wildcharm','wildcharm'),
        ('wildignore','wildignore'),
        ('wildignorecase','wildignorecase'),
        ('wildmenu','wildmenu'),
        ('wildmode','wildmode'),
        ('wildoptions','wildoptions'),
        ('wim','wim'),
        ('winaltkeys','winaltkeys'),
        ('window','window'),
        ('winfixheight','winfixheight'),
        ('winfixwidth','winfixwidth'),
        ('winheight','winheight'),
        ('winminheight','winminheight'),
        ('winminwidth','winminwidth'),
        ('winwidth','winwidth'),
        ('wiv','wiv'),
        ('wiw','wiw'),
        ('wm','wm'),
        ('wmh','wmh'),
        ('wmnu','wmnu'),
        ('wmw','wmw'),
        ('wop','wop'),
        ('wrap','wrap'),
        ('wrapmargin','wrapmargin'),
        ('wrapscan','wrapscan'),
        ('write','write'),
        ('writeany','writeany'),
        ('writebackup','writebackup'),
        ('writedelay','writedelay'),
        ('ws','ws'),
        ('ww','ww'),
    )
    return var
option = _getoption()




from .compat import collections_abc


class DirectedGraph(object):
    """A graph structure with directed edges."""

    def __init__(self):
        self._vertices = set()
        self._forwards = {}  # <key> -> Set[<key>]
        self._backwards = {}  # <key> -> Set[<key>]

    def __iter__(self):
        return iter(self._vertices)

    def __len__(self):
        return len(self._vertices)

    def __contains__(self, key):
        return key in self._vertices

    def copy(self):
        """Return a shallow copy of this graph."""
        other = DirectedGraph()
        other._vertices = set(self._vertices)
        other._forwards = {k: set(v) for k, v in self._forwards.items()}
        other._backwards = {k: set(v) for k, v in self._backwards.items()}
        return other

    def add(self, key):
        """Add a new vertex to the graph."""
        if key in self._vertices:
            raise ValueError("vertex exists")
        self._vertices.add(key)
        self._forwards[key] = set()
        self._backwards[key] = set()

    def remove(self, key):
        """Remove a vertex from the graph, disconnecting all edges from/to it."""
        self._vertices.remove(key)
        for f in self._forwards.pop(key):
            self._backwards[f].remove(key)
        for t in self._backwards.pop(key):
            self._forwards[t].remove(key)

    def connected(self, f, t):
        return f in self._backwards[t] and t in self._forwards[f]

    def connect(self, f, t):
        """Connect two existing vertices.

        Nothing happens if the vertices are already connected.
        """
        if t not in self._vertices:
            raise KeyError(t)
        self._forwards[f].add(t)
        self._backwards[t].add(f)

    def iter_edges(self):
        for f, children in self._forwards.items():
            for t in children:
                yield f, t

    def iter_children(self, key):
        return iter(self._forwards[key])

    def iter_parents(self, key):
        return iter(self._backwards[key])


class _FactoryIterableView(object):
    """Wrap an iterator factory returned by `find_matches()`.

    Calling `iter()` on this class would invoke the underlying iterator
    factory, making it a "collection with ordering" that can be iterated
    through multiple times, but lacks random access methods presented in
    built-in Python sequence types.
    """

    def __init__(self, factory):
        self._factory = factory

    def __repr__(self):
        return "{}({})".format(type(self).__name__, list(self._factory()))

    def __bool__(self):
        try:
            next(self._factory())
        except StopIteration:
            return False
        return True

    __nonzero__ = __bool__  # XXX: Python 2.

    def __iter__(self):
        return self._factory()

    def for_preference(self):
        """Provide an candidate iterable for `get_preference()`"""
        return self._factory()

    def excluding(self, candidates):
        """Create a new instance excluding specified candidates."""

        def factory():
            return (c for c in self._factory() if c not in candidates)

        return type(self)(factory)


class _SequenceIterableView(object):
    """Wrap an iterable returned by find_matches().

    This is essentially just a proxy to the underlying sequence that provides
    the same interface as `_FactoryIterableView`.
    """

    def __init__(self, sequence):
        self._sequence = sequence

    def __repr__(self):
        return "{}({})".format(type(self).__name__, self._sequence)

    def __bool__(self):
        return bool(self._sequence)

    __nonzero__ = __bool__  # XXX: Python 2.

    def __iter__(self):
        return iter(self._sequence)

    def __len__(self):
        return len(self._sequence)

    def for_preference(self):
        """Provide an candidate iterable for `get_preference()`"""
        return self._sequence

    def excluding(self, candidates):
        """Create a new instance excluding specified candidates."""
        return type(self)([c for c in self._sequence if c not in candidates])


def build_iter_view(matches):
    """Build an iterable view from the value returned by `find_matches()`."""
    if callable(matches):
        return _FactoryIterableView(matches)
    if not isinstance(matches, collections_abc.Sequence):
        matches = list(matches)
    return _SequenceIterableView(matches)
